# SPDX-License-Identifier: LGPL-2.1-or-later

import tests.xdp_utils as xdp

import dbus
import pytest


CORRECT_APP_ID_DESKTOP = b"""
[Desktop Entry]
Version=1.0
Name=CorrectAppId
Exec=true %u
Type=Application
"""


@pytest.fixture
def xdg_data_home_files():
    return {
        "applications/org.example.CorrectAppId.desktop": CORRECT_APP_ID_DESKTOP,
    }


@pytest.fixture
def xdp_app_info() -> xdp.AppInfo:
    return xdp.AppInfo.new_host(
        app_id="org.example.WrongAppId",
    )


@pytest.fixture
def required_templates():
    return {"remotedesktop": {}}


class TestRegistry:
    def test_version(self, portals, dbus_con):
        documents = dbus_con.get_object(
            "org.freedesktop.portal.Desktop",
            "/org/freedesktop/portal/desktop",
        )

        properties_intf = dbus.Interface(
            documents,
            "org.freedesktop.DBus.Properties",
        )
        portal_version = properties_intf.Get(
            "org.freedesktop.host.portal.Registry",
            "version",
        )
        assert int(portal_version) == 1

    def create_dummy_session(self, dbus_con):
        remotedesktop_intf = xdp.get_portal_iface(dbus_con, "RemoteDesktop")
        request = xdp.Request(dbus_con, remotedesktop_intf)

        session_counter_attr_name = "session_counter"
        session_counter = getattr(self, session_counter_attr_name, 0)
        setattr(self, session_counter_attr_name, session_counter + 1)

        options = {
            "session_handle_token": f"session_token{session_counter}",
        }
        response = request.call(
            "CreateSession",
            options=options,
        )
        assert response
        assert response.response == 0

        return xdp.Session.from_response(dbus_con, response)

    def test_registerless(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        mock_intf = xdp.get_mock_iface(dbus_con)

        expected_app_id = app_id

        session = self.create_dummy_session(dbus_con)

        app_id = mock_intf.GetSessionAppId(session.handle)
        assert app_id == expected_app_id

    def test_register(self, portals, dbus_con):
        registry_intf = xdp.get_portal_iface(dbus_con, "Registry", domain="host")
        mock_intf = xdp.get_mock_iface(dbus_con)

        expected_app_id = "org.example.CorrectAppId"
        registry_intf.Register(expected_app_id, {})

        session = self.create_dummy_session(dbus_con)

        app_id = mock_intf.GetSessionAppId(session.handle)
        assert app_id == expected_app_id

    def test_late_register(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        registry_intf = xdp.get_portal_iface(dbus_con, "Registry", domain="host")
        mock_intf = xdp.get_mock_iface(dbus_con)

        expected_app_id = app_id
        unexpected_app_id = "org.example.CorrectAppId"

        session = self.create_dummy_session(dbus_con)

        app_id = mock_intf.GetSessionAppId(session.handle)
        assert app_id == expected_app_id

        with pytest.raises(dbus.exceptions.DBusException) as exc_info:
            registry_intf.Register(unexpected_app_id, {})
        exc_info.match(".*Connection already associated with an application ID.*")

        new_session = self.create_dummy_session(dbus_con)

        new_app_id = mock_intf.GetSessionAppId(new_session.handle)
        assert new_app_id == expected_app_id

    def test_multiple_connections(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        registry_intf = xdp.get_portal_iface(dbus_con, "Registry", domain="host")
        mock_intf = xdp.get_mock_iface(dbus_con)

        expected_app_id = "org.example.CorrectAppId"
        unexpected_app_id = app_id

        registry_intf.Register(expected_app_id, {})
        session = self.create_dummy_session(dbus_con)
        app_id = mock_intf.GetSessionAppId(session.handle)
        assert app_id == expected_app_id

        dbus_con2 = dbus.bus.BusConnection(dbus.bus.BusConnection.TYPE_SESSION)
        dbus_con2.set_exit_on_disconnect(False)
        mock_intf2 = xdp.get_mock_iface(dbus_con2)
        session2 = self.create_dummy_session(dbus_con2)
        app_id2 = mock_intf2.GetSessionAppId(session2.handle)
        assert app_id2 == unexpected_app_id
        dbus_con2.close()

        dbus_con3 = dbus.bus.BusConnection(dbus.bus.BusConnection.TYPE_SESSION)
        dbus_con3.set_exit_on_disconnect(False)
        mock_intf3 = xdp.get_mock_iface(dbus_con3)
        registry_intf3 = xdp.get_portal_iface(dbus_con3, "Registry", domain="host")
        registry_intf3.Register(expected_app_id, {})
        session3 = self.create_dummy_session(dbus_con3)
        app_id3 = mock_intf3.GetSessionAppId(session3.handle)
        assert app_id3 == expected_app_id
        dbus_con3.close()

    def test_no_reregister(self, portals, dbus_con):
        registry_intf = xdp.get_portal_iface(dbus_con, "Registry", domain="host")
        mock_intf = xdp.get_mock_iface(dbus_con)

        expected_app_id = "org.example.CorrectAppId"

        registry_intf.Register(expected_app_id, {})
        session = self.create_dummy_session(dbus_con)
        app_id = mock_intf.GetSessionAppId(session.handle)
        assert app_id == expected_app_id

        with pytest.raises(dbus.exceptions.DBusException) as exc_info:
            registry_intf.Register(expected_app_id, {})
        exc_info.match(".*Connection already associated with an application ID.*")



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black
# mypy: disable-error-code="misc"

from tests.templates.xdp_utils import Response, init_logger, ImplRequest, ImplSession

import dbus
import dbus.service
import time
from dbusmock import MOCK_IFACE
from gi.repository import GLib
from dataclasses import dataclass


BUS_NAME = "org.freedesktop.impl.portal.Test"
MAIN_OBJ = "/org/freedesktop/portal/desktop"
SYSTEM_BUS = False
MAIN_IFACE = "org.freedesktop.impl.portal.GlobalShortcuts"
VERSION = 1


logger = init_logger(__name__)


@dataclass
class GlobalshortcutsParameters:
    delay: int
    response: int
    expect_close: bool
    force_close: int


def load(mock, parameters={}):
    logger.debug(f"Loading parameters: {parameters}")

    assert not hasattr(mock, "globalshortcuts_params")
    mock.globalshortcuts_params = GlobalshortcutsParameters(
        delay=parameters.get("delay", 200),
        response=parameters.get("response", 0),
        expect_close=parameters.get("expect-close", False),
        force_close=parameters.get("force-close", 0),
    )

    mock.AddProperties(
        MAIN_IFACE,
        dbus.Dictionary(
            {
                "version": dbus.UInt32(parameters.get("version", VERSION)),
            }
        ),
    )
    mock.sessions: dict[str, ImplSession] = {}


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oosa{sv}",
    out_signature="ua{sv}",
    async_callbacks=("cb_success", "cb_error"),
)
def CreateSession(self, handle, session_handle, app_id, options, cb_success, cb_error):
    logger.debug(f"CreateSession({handle}, {session_handle}, {app_id}, {options})")
    params = self.globalshortcuts_params

    session = ImplSession(self, BUS_NAME, session_handle, app_id).export()
    self.sessions[session_handle] = session

    request = ImplRequest(
        self,
        BUS_NAME,
        handle,
        logger,
        cb_success,
        cb_error,
    )

    if params.expect_close:
        request.wait_for_close()
    else:
        request.respond(
            Response(params.response, {"session_handle": session.handle}),
            delay=params.delay,
        )
        if params.force_close > 0:
            GLib.timeout_add(params.force_close, session.close)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="ooa(sa{sv})sa{sv}",
    out_signature="ua{sv}",
    async_callbacks=("cb_success", "cb_error"),
)
def BindShortcuts(
    self,
    handle,
    session_handle,
    shortcuts,
    parent_window,
    options,
    cb_success,
    cb_error,
):
    logger.debug(f"BindShortcuts({handle}, {session_handle}, {shortcuts}, {options})")
    params = self.globalshortcuts_params

    assert session_handle in self.sessions

    request = ImplRequest(
        self,
        BUS_NAME,
        handle,
        logger,
        cb_success,
        cb_error,
    )

    if params.expect_close:
        request.wait_for_close()
    else:

        def reply():
            logger.debug(f"BindShortcuts with shortcuts {shortcuts}")
            self.sessions[session_handle].shortcuts = shortcuts
            return Response(params.response, {})

        request.respond(reply, delay=params.delay)


@dbus.service.method(
    MAIN_IFACE,
    in_signature="oo",
    out_signature="ua{sv}",
)
def ListShortcuts(
    self,
    handle,
    session_handle,
):
    shortcuts = self.sessions[session_handle].shortcuts
    return (0, {"shortcuts": shortcuts})


@dbus.service.method(
    MAIN_IFACE,
    in_signature="osa{sv}",
    out_signature="",
)
def ConfigureShortcuts(self, session_handle, parent_window, options):
    assert session_handle in self.sessions


@dbus.service.method(
    MOCK_IFACE,
    in_signature="os",
    out_signature="",
)
def Trigger(self, session_handle, shortcut_id):
    now_since_epoch = int(time.time() * 1000000)
    self.EmitSignal(
        MAIN_IFACE,
        "Activated",
        "osta{sv}",
        [session_handle, shortcut_id, now_since_epoch, {}],
    )
    time.sleep(0.2)
    now_since_epoch = int(time.time() * 1000000)
    self.EmitSignal(
        MAIN_IFACE,
        "Deactivated",
        "osta{sv}",
        [session_handle, shortcut_id, now_since_epoch, {}],
    )



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import dbus
import pytest
import socket
from typing import List, Dict, Any


@pytest.fixture
def required_templates():
    return {"remotedesktop": {}}


class TestRemoteDesktop:
    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "RemoteDesktop", 2)

    def test_create_close_session(self, portals, dbus_con):
        remotedesktop_intf = xdp.get_portal_iface(dbus_con, "RemoteDesktop")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {
            "session_handle_token": "session_token0",
        }
        response = request.call(
            "CreateSession",
            options=options,
        )

        assert response
        assert response.response == 0

        session = xdp.Session.from_response(dbus_con, response)
        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("CreateSession")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == session.handle
        # assert args[2] == ""  # appid, not necessary empty

        session.close()
        xdp.wait_for(lambda: session.closed)

    @pytest.mark.parametrize(
        "token", ("Invalid-Token&", "", "/foo", "something-else", "😄")
    )
    def test_remote_desktop_create_session_invalid(self, portals, dbus_con, token):
        remotedesktop_intf = xdp.get_portal_iface(dbus_con, "RemoteDesktop")

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {"session_handle_token": token}

        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            request.call("CreateSession", options=options)

        e = excinfo.value
        assert e.get_dbus_name() == "org.freedesktop.portal.Error.InvalidArgument"
        assert "Invalid token" in e.get_dbus_message()

    @pytest.mark.parametrize(
        "template_params", ({"remotedesktop": {"force-close": 500}},)
    )
    def test_create_session_signal_closed(self, portals, dbus_con):
        remotedesktop_intf = xdp.get_portal_iface(dbus_con, "RemoteDesktop")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {
            "session_handle_token": "session_token0",
        }
        response = request.call(
            "CreateSession",
            options=options,
        )

        assert response
        assert response.response == 0

        session = xdp.Session.from_response(dbus_con, response)
        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("CreateSession")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == session.handle
        # assert args[2] == ""  # appid, not necessary empty

        # Now expect the backend to close it
        xdp.wait_for(lambda: session.closed)

    def test_connect_to_eis(self, portals, dbus_con):
        remotedesktop_intf = xdp.get_portal_iface(dbus_con, "RemoteDesktop")

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {
            "session_handle_token": "session_token0",
        }
        response = request.call(
            "CreateSession",
            options=options,
        )

        assert response
        assert response.response == 0

        session = xdp.Session.from_response(dbus_con, response)
        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {
            "types": dbus.UInt32(0x3),
        }
        response = request.call(
            "SelectDevices",
            session_handle=session.handle,
            options=options,
        )
        assert response
        assert response.response == 0

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {}
        response = request.call(
            "Start",
            session_handle=session.handle,
            parent_window="",
            options=options,
        )
        assert response
        assert response.response == 0

        fd = remotedesktop_intf.ConnectToEIS(
            session.handle,
            dbus.Dictionary({}, signature="sv"),
        )
        eis_socket = socket.fromfd(fd.take(), socket.AF_UNIX, socket.SOCK_STREAM)
        assert eis_socket.recv(10) == b"HELLO"

    @pytest.mark.parametrize(
        "template_params", ({"remotedesktop": {"fail-connect-to-eis": True}},)
    )
    def test_connect_to_eis_fail(self, portals, dbus_con):
        remotedesktop_intf = xdp.get_portal_iface(dbus_con, "RemoteDesktop")

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {
            "session_handle_token": "session_token0",
        }
        response = request.call(
            "CreateSession",
            options=options,
        )

        assert response
        assert response.response == 0

        session = xdp.Session.from_response(dbus_con, response)
        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {
            "types": dbus.UInt32(0x3),
        }
        response = request.call(
            "SelectDevices",
            session_handle=session.handle,
            options=options,
        )
        assert response
        assert response.response == 0

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {}
        response = request.call(
            "Start",
            session_handle=session.handle,
            parent_window="",
            options=options,
        )
        assert response
        assert response.response == 0

        with pytest.raises(dbus.exceptions.DBusException) as excinfo:
            _ = remotedesktop_intf.ConnectToEIS(
                session.handle, dbus.Dictionary({}, signature="sv")
            )
        assert "Purposely failing ConnectToEIS" in excinfo.value.get_dbus_message()

    def test_connect_to_eis_fail_notifies(self, portals, dbus_con):
        remotedesktop_intf = xdp.get_portal_iface(dbus_con, "RemoteDesktop")

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {
            "session_handle_token": "session_token0",
        }
        response = request.call(
            "CreateSession",
            options=options,
        )

        assert response
        assert response.response == 0

        session = xdp.Session.from_response(dbus_con, response)
        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {
            "types": dbus.UInt32(0x3),
        }
        response = request.call(
            "SelectDevices",
            session_handle=session.handle,
            options=options,
        )
        assert response
        assert response.response == 0

        request = xdp.Request(dbus_con, remotedesktop_intf)
        options = {}
        response = request.call(
            "Start",
            session_handle=session.handle,
            parent_window="",
            options=options,
        )
        assert response
        assert response.response == 0

        notifyfuncs: List[Dict[str, Any]] = [
            {"name": "NotifyPointerMotion", "args": (1, 2)},
            {"name": "NotifyPointerMotionAbsolute", "args": (0, 1, 2)},
            {"name": "NotifyPointerButton", "args": (1, 1)},
            {"name": "NotifyPointerAxis", "args": (1, 1)},
            {"name": "NotifyPointerAxisDiscrete", "args": (1, 1)},
            {"name": "NotifyKeyboardKeycode", "args": (1, 1)},
            {"name": "NotifyKeyboardKeysym", "args": (1, 1)},
            {"name": "NotifyTouchDown", "args": (0, 0, 1, 1)},
            {"name": "NotifyTouchMotion", "args": (0, 0, 1, 1)},
            {"name": "NotifyTouchUp", "args": (0,)},
        ]
        for notifyfunc in notifyfuncs:
            with pytest.raises(dbus.exceptions.DBusException) as excinfo:
                func = getattr(remotedesktop_intf, notifyfunc["name"])
                assert func is not None
                func(
                    session.handle,
                    dbus.Dictionary({}, signature="sv"),
                    *notifyfunc["args"],
                )
            # Not the best error message but...
            assert (
                "Session is not allowed to call Notify"
                in excinfo.value.get_dbus_message()
            )



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import dbus
import pytest
from typing import Any
import os
from pathlib import Path


SCREENSHOT_DATA = dbus.Dictionary(
    {
        "uri": "file:///screenshot.png",
        "color": (0.0, 1.0, 0.331),
    },
    signature="sv",
)


@pytest.fixture
def required_templates():
    image = Path(os.environ["XDG_DATA_HOME"]) / "screenshot-image.png"
    image.write_text("image contents")
    SCREENSHOT_DATA["uri"] = f"file://{image.absolute().as_posix()}"

    return {
        "access": {},
        "screenshot": {
            "results": SCREENSHOT_DATA,
        },
    }


class TestScreenshot:
    def set_permission(self, dbus_con, appid, permission):
        perm_store_intf = xdp.get_permission_store_iface(dbus_con)
        perm_store_intf.SetPermission(
            "screenshot",
            True,
            "screenshot",
            appid,
            [permission],
        )

    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Screenshot", 2)

    @pytest.mark.parametrize("modal", [True, False])
    @pytest.mark.parametrize("interactive", [True, False])
    def test_screenshot_basic(
        self, xdg_document_portal, portals, dbus_con, xdp_app_info, modal, interactive
    ):
        app_id = xdp_app_info.app_id
        screenshot_intf = xdp.get_portal_iface(dbus_con, "Screenshot")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, screenshot_intf)
        options = {
            "modal": modal,
            "interactive": interactive,
        }
        response = request.call(
            "Screenshot",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 0

        assert xdp.uri_same_file(SCREENSHOT_DATA["uri"], response.results["uri"])

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Screenshot")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3]["modal"] == modal
        assert args[3]["interactive"] == interactive

        # check that args were forwarded to access portal correctly
        if not interactive:
            method_calls = mock_intf.GetMethodCalls("AccessDialog")
            assert len(method_calls) > 0
            _, args = method_calls[-1]
            assert args[1] == app_id
            assert args[2] == ""  # parent window
            assert args[6]["modal"] == modal

    @pytest.mark.parametrize(
        "template_params", ({"screenshot": {"expect-close": True}},)
    )
    def test_screenshot_close(self, portals, dbus_con):
        screenshot_intf = xdp.get_portal_iface(dbus_con, "Screenshot")

        request = xdp.Request(dbus_con, screenshot_intf)
        request.schedule_close(1000)
        options = {
            "interactive": True,
        }
        request.call(
            "Screenshot",
            parent_window="",
            options=options,
        )

        # Only true if the impl.Request was closed too
        assert request.closed

    @pytest.mark.parametrize("template_params", ({"screenshot": {"response": 1}},))
    def test_screenshot_cancel(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        screenshot_intf = xdp.get_portal_iface(dbus_con, "Screenshot")
        mock_intf = xdp.get_mock_iface(dbus_con)

        modal = True
        interactive = True

        request = xdp.Request(dbus_con, screenshot_intf)
        options = {
            "modal": modal,
            "interactive": interactive,
        }
        response = request.call(
            "Screenshot",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 1

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Screenshot")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == app_id
        assert args[2] == ""  # parent window
        assert args[3]["modal"] == modal
        assert args[3]["interactive"] == interactive

    @pytest.mark.parametrize("permission", ["", "yes", "no", "ask"])
    def test_screenshot_permissions(
        self, xdg_document_portal, portals, dbus_con, xdp_app_info, permission
    ):
        app_id = xdp_app_info.app_id
        screenshot_intf = xdp.get_portal_iface(dbus_con, "Screenshot")
        mock_intf = xdp.get_mock_iface(dbus_con)

        self.set_permission(dbus_con, app_id, permission)

        request = xdp.Request(dbus_con, screenshot_intf)
        options = {
            "modal": True,
            "interactive": False,
            "handle_token": request.handle_token,
        }
        response = request.call(
            "Screenshot",
            parent_window="",
            options=options,
        )

        assert response
        if permission == "no":
            assert response.response != 0
            return

        assert response.response == 0
        method_calls = mock_intf.GetMethodCalls("AccessDialog")
        if permission == "yes":
            assert len(method_calls) == 0
        else:
            assert len(method_calls) > 0

        request = xdp.Request(dbus_con, screenshot_intf)
        options["handle_token"] = request.handle_token
        response = request.call(
            "Screenshot",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 0
        method_calls_2 = mock_intf.GetMethodCalls("AccessDialog")
        if permission == "ask":
            assert len(method_calls_2) > len(method_calls)
        else:
            assert len(method_calls_2) == len(method_calls)

    def test_pick_color_basic(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        screenshot_intf = xdp.get_portal_iface(dbus_con, "Screenshot")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, screenshot_intf)
        options: Any = {}
        response = request.call(
            "PickColor",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["color"] == SCREENSHOT_DATA["color"]

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("PickColor")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == app_id
        assert args[2] == ""  # parent window

    @pytest.mark.parametrize(
        "template_params", ({"screenshot": {"expect-close": True}},)
    )
    def test_pick_color_close(self, portals, dbus_con):
        screenshot_intf = xdp.get_portal_iface(dbus_con, "Screenshot")

        request = xdp.Request(dbus_con, screenshot_intf)
        request.schedule_close(1000)
        options: Any = {}
        request.call(
            "PickColor",
            parent_window="",
            options=options,
        )

        # Only true if the impl.Request was closed too
        assert request.closed

    @pytest.mark.parametrize("template_params", ({"screenshot": {"response": 1}},))
    def test_pick_color_cancel(self, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        screenshot_intf = xdp.get_portal_iface(dbus_con, "Screenshot")
        mock_intf = xdp.get_mock_iface(dbus_con)

        request = xdp.Request(dbus_con, screenshot_intf)
        options: Any = {}
        response = request.call(
            "PickColor",
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 1

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("PickColor")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == app_id
        assert args[2] == ""  # parent window



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp
import tests.test_inputcapture as inputcapture

import dbus
import pytest
import os
from enum import Enum


@pytest.fixture
def required_templates():
    return {
        "Clipboard": {},
        "RemoteDesktop": {"force-clipboard-enabled": True},
        "InputCapture": {
            "force-clipboard-enabled": True,
            "activated-delay": 100,
        },
    }


class SessionType(Enum):
    REMOTE_DESKTOP = 1
    INPUT_CAPTURE = 2


@pytest.mark.parametrize(
    "type", (SessionType.REMOTE_DESKTOP, SessionType.INPUT_CAPTURE)
)
class TestClipboard:
    def test_version(self, portals, dbus_con, type):
        xdp.check_version(dbus_con, "Clipboard", 1)

    def start_remote_desktop_session(self, dbus_con):
        clipboard_intf = xdp.get_portal_iface(dbus_con, "Clipboard")
        remotedesktop_intf = xdp.get_portal_iface(dbus_con, "RemoteDesktop")

        create_session_request = xdp.Request(dbus_con, remotedesktop_intf)
        create_session_response = create_session_request.call(
            "CreateSession", options={"session_handle_token": "1234"}
        )
        assert create_session_response
        assert create_session_response.response == 0
        assert str(create_session_response.results["session_handle"])

        session = xdp.Session.from_response(dbus_con, create_session_response)

        clipboard_intf.RequestClipboard(session.handle, {})

        start_session_request = xdp.Request(dbus_con, remotedesktop_intf)
        start_session_response = start_session_request.call(
            "Start", session_handle=session.handle, parent_window="", options={}
        )

        assert start_session_response
        assert start_session_response.response == 0

        return (session, start_session_response.results.get("clipboard_enabled"))

    def start_input_capture_session(self, dbus_con):
        inputcapture_intf = xdp.get_portal_iface(dbus_con, "InputCapture")

        session = inputcapture.InputcaptureSession(dbus_con)
        session.create()
        start_session_response = session.start()
        session.get_zones()
        # The default zone is 1920x1080
        barriers = [
            {
                "barrier_id": dbus.UInt32(10, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1920, 0], signature="iiii", variant_level=1
                ),
            },
        ]
        session.set_pointer_barriers(barriers)
        session.connect_to_eis()

        activated_signal_received = False

        def cb_activated(session_handle, options):
            nonlocal activated_signal_received
            activated_signal_received = True

        inputcapture_intf.connect_to_signal("Activated", cb_activated)

        session.enable()

        xdp.wait_for(lambda: activated_signal_received)

        return (
            session.session,
            start_session_response.results.get("clipboard_enabled"),
        )

    def start_session(self, dbus_con, type):
        if type == SessionType.REMOTE_DESKTOP:
            return self.start_remote_desktop_session(dbus_con)
        if type == SessionType.INPUT_CAPTURE:
            return self.start_input_capture_session(dbus_con)
        raise Exception("Unknown type")

    def test_request_clipboard_and_start_session(self, portals, dbus_con, type):
        _, clipboard_enabled = self.start_session(dbus_con, type)

        assert clipboard_enabled

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "RemoteDesktop": {"force-clipboard-enabled": False},
                "InputCapture": {
                    "force-clipboard-enabled": False,
                    "activated-delay": 100,
                },
            },
        ),
    )
    def test_checks_clipboard_enabled(self, portals, dbus_con, type):
        clipboard_intf = xdp.get_portal_iface(dbus_con, "Clipboard")
        session, clipboard_enabled = self.start_session(dbus_con, type)

        assert not clipboard_enabled

        with pytest.raises(dbus.exceptions.DBusException):
            clipboard_intf.SetSelection(session.handle, {})

    def test_set_selection(self, portals, dbus_con, type):
        clipboard_intf = xdp.get_portal_iface(dbus_con, "Clipboard")
        session, _ = self.start_session(dbus_con, type)

        clipboard_intf.SetSelection(session.handle, {})

    def test_selection_write(self, portals, dbus_con, type):
        clipboard_intf = xdp.get_portal_iface(dbus_con, "Clipboard")
        session, _ = self.start_session(dbus_con, type)

        fd_object: dbus.types.UnixFd = clipboard_intf.SelectionWrite(
            session.handle, 1234
        )
        assert fd_object

        fd = fd_object.take()
        assert fd

        bytes_written = os.write(fd, b"Clipboard")
        assert bytes_written > 0

        clipboard_intf.SelectionWriteDone(session.handle, 1234, True)

    def test_selection_read(self, portals, dbus_con, type):
        clipboard_intf = xdp.get_portal_iface(dbus_con, "Clipboard")
        session, _ = self.start_session(dbus_con, type)

        fd_object: dbus.types.UnixFd = clipboard_intf.SelectionRead(
            session.handle, "mimetype"
        )
        assert fd_object

        fd = fd_object.take()
        assert fd

        clipboard_contents = os.read(fd, 1000)
        assert str(clipboard_contents)



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

from dbus.mainloop.glib import DBusGMainLoop
from gi.repository import GLib, Gio, GioUnix
from itertools import count
from typing import Any, Dict, Optional, NamedTuple, Callable, List
from pathlib import Path
from enum import Enum, IntEnum
from dataclasses import dataclass, field
from urllib.parse import unquote, urlparse

import os
import dbus
import dbus.proxies
import dbusmock
import logging
import subprocess


DBusGMainLoop(set_as_default=True)

# Anything that takes longer than 5s needs to fail
DBUS_TIMEOUT = int(os.environ.get("XDP_DBUS_TIMEOUT", "5000"))

_counter = count()

ASV = Dict[str, Any]


def init_logger(name: str) -> logging.Logger:
    """
    Common logging setup for tests. Use as:

        >>> import tests.xdp_utils as xdp
        >>> logger = xdp.init_logger(__name__)
        >>> logger.debug("foo")

    """
    logging.basicConfig(
        format="%(levelname).1s|%(name)s: %(message)s", level=logging.DEBUG
    )
    logger = logging.getLogger(f"xdp.{name}")
    logger.setLevel(logging.DEBUG)
    return logger


logger = init_logger("utils")


def is_in_ci() -> bool:
    return os.environ.get("XDP_TEST_IN_CI") is not None


def is_in_container() -> bool:
    return is_in_ci() or (
        "container" in os.environ
        and (os.environ["container"] == "docker" or os.environ["container"] == "podman")
    )


def run_long_tests() -> bool:
    return os.environ.get("XDP_TEST_RUN_LONG") is not None


def is_valgrind() -> bool:
    return os.getenv("XDP_TEST_VALGRIND") is not None


def check_program_success(cmd) -> bool:
    proc = subprocess.Popen(
        cmd, stdout=None, stderr=None, shell=True, universal_newlines=True
    )
    _ = proc.communicate()
    return proc.returncode == 0


def uri_same_file(uri1, uri2):
    orig = Path(unquote(urlparse(uri1).path))
    path = Path(unquote(urlparse(uri2).path))
    return orig.read_text() == path.read_text()


def uris_same_files(uris, uris_other):
    return all(
        uri_same_file(uri, uri_other) for uri, uri_other in zip(uris, uris_other)
    )


class FuseNotSupportedException(Exception):
    pass


def ensure_fuse_supported() -> None:
    if not check_program_success("fusermount3 --version"):
        raise FuseNotSupportedException("no fusermount3")

    if not check_program_success(
        "capsh --print | grep -q 'Bounding set.*[^a-z]cap_sys_admin'"
    ):
        raise FuseNotSupportedException(
            "No cap_sys_admin in bounding set, can't use FUSE"
        )

    if not check_program_success("[ -w /dev/fuse ]"):
        raise FuseNotSupportedException("no write access to /dev/fuse")

    if not check_program_success("[ -e /etc/mtab ]"):
        raise FuseNotSupportedException("no /etc/mtab")


def wait(ms: int):
    """
    Waits for the specified amount of milliseconds.
    """
    mainloop = GLib.MainLoop()
    GLib.timeout_add(ms, mainloop.quit)
    mainloop.run()


def wait_for(fn: Callable[[], bool]):
    """
    Waits and dispatches to mainloop until the function fn returns true. This is
    useful in combination with a lambda which captures a variable:

        my_var = False
        def callback():
            my_var = True
        do_something_later(callback)
        xdp.wait_for(lambda: my_var)
    """
    mainloop = GLib.MainLoop()
    while not fn():
        GLib.timeout_add(50, mainloop.quit)
        mainloop.run()


def get_permission_store_iface(bus: dbus.Bus) -> dbus.Interface:
    """
    Returns the dbus interface of the xdg-permission-store.
    """
    obj = bus.get_object(
        "org.freedesktop.impl.portal.PermissionStore",
        "/org/freedesktop/impl/portal/PermissionStore",
    )
    return dbus.Interface(obj, "org.freedesktop.impl.portal.PermissionStore")


def get_document_portal_iface(bus: dbus.Bus) -> dbus.Interface:
    """
    Returns the dbus interface of the xdg-document-portal.
    """
    obj = bus.get_object(
        "org.freedesktop.portal.Documents",
        "/org/freedesktop/portal/documents",
    )
    return dbus.Interface(obj, "org.freedesktop.portal.Documents")


def get_mock_iface(bus: dbus.Bus, bus_name: Optional[str] = None) -> dbus.Interface:
    """
    Returns the mock interface of the xdg-desktop-portal.
    """
    if not bus_name:
        bus_name = "org.freedesktop.impl.portal.Test"

    obj = bus.get_object(bus_name, "/org/freedesktop/portal/desktop")
    return dbus.Interface(obj, dbusmock.MOCK_IFACE)


def portal_interface_name(portal_name: str, domain: Optional[str] = None) -> str:
    """
    Returns the fully qualified interface for a portal name.
    """
    if domain:
        return f"org.freedesktop.{domain}.portal.{portal_name}"
    else:
        return f"org.freedesktop.portal.{portal_name}"


def get_portal_iface(
    bus: dbus.Bus, name: str, domain: Optional[str] = None
) -> dbus.Interface:
    """
    Returns the dbus interface for a portal name.
    """
    name = portal_interface_name(name, domain)
    return get_iface(bus, name)


def get_iface(bus: dbus.Bus, name: str) -> dbus.Interface:
    """
    Returns a named interface of the main portal object.
    """
    try:
        ifaces = bus._xdp_portal_ifaces
    except AttributeError:
        ifaces = bus._xdp_portal_ifaces = {}

    try:
        intf = ifaces[name]
    except KeyError:
        intf = dbus.Interface(get_xdp_dbus_object(bus), name)
        assert intf
        ifaces[name] = intf
    return intf


def get_xdp_dbus_object(bus: dbus.Bus) -> dbus.proxies.ProxyObject:
    """
    Returns the main portal object.
    """
    try:
        obj = getattr(bus, "_xdp_dbus_object")
    except AttributeError:
        obj = bus.get_object(
            "org.freedesktop.portal.Desktop", "/org/freedesktop/portal/desktop"
        )
        assert obj
        bus._xdp_dbus_object = obj
    return obj


def check_version(bus: dbus.Bus, portal_name: str, expected_version: int):
    """
    Checks that the portal_name portal version is equal to expected_version.
    """
    properties_intf = dbus.Interface(
        get_xdp_dbus_object(bus), "org.freedesktop.DBus.Properties"
    )
    portal_iface_name = portal_interface_name(portal_name)
    try:
        portal_version = properties_intf.Get(portal_iface_name, "version")
        assert int(portal_version) == expected_version
    except dbus.exceptions.DBusException as e:
        logger.critical(e)
        assert e is None, str(e)


def desktop_files_path() -> Path:
    """Returns the default path for desktop files"""
    return Path(os.environ["XDG_DATA_HOME"]) / "applications"


class AppInfoKind(Enum):
    HOST = 1
    FLATPAK = 2
    SNAP = 3
    LINYAPS = 4


@dataclass
class AppInfo:
    """
    Interacts with conftest.py via ensure_files and extend_env to make the
    portal frontend discover the requested XdpAppInfo for incoming connections.

    Testing code can use this class to construct a specific XdpAppInfo for the
    xdp_app_info fixture.
    """

    kind: AppInfoKind
    app_id: str
    desktop_file: str
    env: dict[str, str] = field(default_factory=dict)
    files: dict[Path, bytes] = field(default_factory=dict)

    def ensure_files(self) -> None:
        for path, content in self.files.items():
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_bytes(content)

    def extend_env(self, env: dict[str, str]) -> None:
        for key, val in self.env.items():
            env[key] = val

    def gapp_info(self) -> GioUnix.DesktopAppInfo:
        desktop_file_path = desktop_files_path() / self.desktop_file
        if not os.path.exists(desktop_file_path):
            return None

        return GioUnix.DesktopAppInfo.new_from_filename(str(desktop_file_path))

    @classmethod
    def new_host(
        cls,
        app_id: str,
        desktop_entry: bytes | None = None,
    ):
        kind = AppInfoKind.HOST
        desktop_file = f"{app_id}.desktop"
        env = {
            "XDG_DESKTOP_PORTAL_TEST_APP_INFO_KIND": "host",
            "XDG_DESKTOP_PORTAL_TEST_HOST_APPID": app_id,
        }
        files = {}

        if desktop_entry:
            files[desktop_files_path() / desktop_file] = desktop_entry

        return cls(
            kind=kind,
            app_id=app_id,
            desktop_file=desktop_file,
            env=env,
            files=files,
        )

    @classmethod
    def new_flatpak(
        cls,
        app_id: str,
        instance_id: str | None = None,
        usb_queries: str | None = None,
        desktop_entry: bytes | None = None,
        metadata: bytes | None = None,
    ):
        kind = AppInfoKind.FLATPAK
        desktop_file = f"{app_id}.desktop"
        env = {
            "XDG_DESKTOP_PORTAL_TEST_APP_INFO_KIND": "flatpak",
        }
        files = {}

        if not instance_id:
            instance_id = "1234567890"

        if not desktop_entry:
            desktop_entry = b"""
[Desktop Entry]
Version=1.0
Name=Example App
Exec=true %u
Type=Application
"""

        files[desktop_files_path() / desktop_file] = desktop_entry

        if not metadata:
            metadata_str = f"""
[Application]
name={app_id}
runtime=org.freedesktop.Platform/x86_64/23.08
sdk=org.freedesktop.Sdk/x86_64/23.08
command={app_id}

[Instance]
instance-id={instance_id}

[Context]
shared=network;ipc;
sockets=x11;wayland;pulseaudio;fallback-x11;
devices=dri;
"""
            metadata = metadata_str.encode("utf8")

        if usb_queries:
            metadata_usb_str = f"""
[USB Devices]
enumerable-devices={usb_queries}
"""
            metadata += metadata_usb_str.encode("utf8")

        metadata_path = Path(os.environ["TMPDIR"]) / "flatpak-metadata"

        files[metadata_path] = metadata
        env["XDG_DESKTOP_PORTAL_TEST_FLATPAK_METADATA"] = (
            metadata_path.absolute().as_posix()
        )

        return cls(
            kind=kind,
            app_id=app_id,
            desktop_file=desktop_file,
            env=env,
            files=files,
        )

    @classmethod
    def new_snap(
        cls,
        common_id: str,
        snap_name: str,
        app_name: str,
        desktop_entry: bytes | None = None,
        metadata: bytes | None = None,
    ):
        kind = AppInfoKind.SNAP
        app_id = f"snap.{snap_name}"
        desktop_file = f"{snap_name}_{app_name}.desktop"
        env = {
            "XDG_DESKTOP_PORTAL_TEST_APP_INFO_KIND": "snap",
        }
        files = {}

        if not desktop_entry:
            desktop_entry_str = f"""
[Desktop Entry]
Version=1.0
Name=Example App
Exec=true %u
Type=Application
X-SnapInstanceName={snap_name}
X-SnapAppName={app_name}
"""
            desktop_entry = desktop_entry_str.encode("UTF-8")

        files[desktop_files_path() / desktop_file] = desktop_entry

        if not metadata:
            metadata_str = f"""
[Snap Info]
InstanceName={snap_name}
AppName={app_name}
CommonID={common_id}
DesktopFile={desktop_file}
"""
            metadata = metadata_str.encode("UTF-8")

        metadata_path = Path(os.environ["TMPDIR"]) / "snap-metadata"

        files[metadata_path] = metadata
        env["XDG_DESKTOP_PORTAL_TEST_SNAP_METADATA"] = (
            metadata_path.absolute().as_posix()
        )

        return cls(
            kind=kind,
            app_id=app_id,
            desktop_file=desktop_file,
            env=env,
            files=files,
        )

    @classmethod
    def new_linyaps(
        cls,
        app_id: str,
        instance_id: str | None = None,
        desktop_entry: bytes | None = None,
        metadata: bytes | None = None,
    ):
        kind = AppInfoKind.LINYAPS
        desktop_file = f"{app_id}.desktop"
        env = {
            "XDG_DESKTOP_PORTAL_TEST_APP_INFO_KIND": "linyaps",
        }
        files = {}

        if not instance_id:
            instance_id = (
                "278575aac695dafe08974feb55c84bba69e862216e980b7ede28c5844e93682c"
            )

        if not desktop_entry:
            desktop_entry = b"""
[Desktop Entry]
Version=1.0
Name=Example App
Exec=true %u
Type=Application
"""

        files[desktop_files_path() / desktop_file] = desktop_entry

        if not metadata:
            metadata_str = f"""
[General]
Linyaps-version=1.10.0

[Application]
Id={app_id}

[Instance]
Id={instance_id}

[Context]
Network=shared
"""
            metadata = metadata_str.encode("utf8")

        metadata_path = Path(os.environ["TMPDIR"]) / "linyaps-metadata"

        files[metadata_path] = metadata
        env["XDG_DESKTOP_PORTAL_TEST_LINYAPS_METADATA"] = (
            metadata_path.absolute().as_posix()
        )

        return cls(
            kind=kind,
            app_id=app_id,
            desktop_file=desktop_file,
            env=env,
            files=files,
        )


class Response(NamedTuple):
    """
    Response as returned by a completed :class:`Request`
    """

    response: int
    results: ASV


class ResponseTimeout(Exception):
    """
    Exception raised by :meth:`Request.call` if the Request did not receive a
    Response in time.
    """

    pass


class Closable:
    """
    Parent class for both Session and Request. Both of these have a Close()
    method.
    """

    def __init__(self, bus: dbus.Bus, objpath: str):
        self.objpath = objpath
        # GLib makes assertions in callbacks impossible, so we wrap all
        # callbacks into a try: except and store the error on the request to
        # be raised later when we're back in the main context
        self.error: Optional[Exception] = None

        self._mainloop: Optional[GLib.MainLoop] = None
        self._impl_closed = False
        self._bus = bus

        self._closable = type(self).__name__
        assert self._closable in ("Request", "Session")
        proxy = bus.get_object("org.freedesktop.portal.Desktop", objpath)
        self._closable_interface = dbus.Interface(
            proxy, f"org.freedesktop.portal.{self._closable}"
        )

    @property
    def bus(self) -> dbus.Bus:
        return self._bus

    @property
    def closed(self) -> bool:
        """
        True if the impl.portal was closed
        """
        return self._impl_closed

    def close(self) -> None:
        signal_match = None

        def cb_impl_closed_by_portal(handle) -> None:
            if handle == self.objpath:
                logger.debug(f"Impl{self._closable} {self.objpath} was closed")
                signal_match.remove()  # type: ignore
                self._impl_closed = True
                if self.closed and self._mainloop:
                    self._mainloop.quit()

        # See :class:`ImplRequest`, this signal is a side-channel for the
        # impl.portal template to notify us when the impl.Request was really
        # closed by the portal.
        signal_match = self._bus.add_signal_receiver(
            cb_impl_closed_by_portal,
            f"{self._closable}Closed",
            dbus_interface="org.freedesktop.impl.portal.Mock",
        )

        logger.debug(f"Closing {self._closable} {self.objpath}")
        self._closable_interface.Close()

    def schedule_close(self, timeout_ms=300):
        """
        Schedule an automatic Close() on the given timeout in milliseconds.
        """
        assert 0 < timeout_ms < DBUS_TIMEOUT
        GLib.timeout_add(timeout_ms * 10 if is_valgrind() else timeout_ms, self.close)


class Request(Closable):
    """
    Helper class for executing methods that use Requests. This calls takes
    care of subscribing to the signals and invokes the method on the
    interface with the expected behaviors. A typical invocation is:

            >>> response = Request(connection, interface).call("Foo", bar="bar")
            >>> assert response.response == 0

    Requests can only be used once, to call a second method you must
    instantiate a new Request object.
    """

    def __init__(self, bus: dbus.Bus, interface: dbus.Interface):
        def sanitize(name):
            return name.lstrip(":").replace(".", "_")

        sender_token = sanitize(bus.get_unique_name())
        self._handle_token = f"request{next(_counter)}"
        self.handle = f"/org/freedesktop/portal/desktop/request/{sender_token}/{self._handle_token}"
        # The Closable
        super().__init__(bus, self.handle)

        self.interface = interface
        self.response: Optional[Response] = None
        self.used = False
        # GLib makes assertions in callbacks impossible, so we wrap all
        # callbacks into a try: except and store the error on the request to
        # be raised later when we're back in the main context
        self.error: Optional[Exception] = None

        proxy = bus.get_object("org.freedesktop.portal.Desktop", self.handle)
        self.mock_interface = dbus.Interface(proxy, dbusmock.MOCK_IFACE)
        self._proxy = bus.get_object("org.freedesktop.portal.Desktop", self.handle)

        def cb_response(response: int, results: ASV) -> None:
            try:
                logger.debug(f"Response received on {self.handle}")
                assert self.response is None
                self.response = Response(response, results)
                if self._mainloop:
                    self._mainloop.quit()
            except Exception as e:
                self.error = e

        self.request_interface = dbus.Interface(proxy, "org.freedesktop.portal.Request")
        self.request_interface.connect_to_signal("Response", cb_response)

    @property
    def handle_token(self) -> dbus.String:
        """
        Returns the dbus-ready handle_token, ready to be put into the options
        """
        return dbus.String(self._handle_token, variant_level=1)

    def call(self, methodname: str, **kwargs) -> Optional[Response]:
        """
        Semi-synchronously call method ``methodname`` on the interface given
        in the Request's constructor. The kwargs must be specified in the
        order the DBus method takes them but the handle_token is automatically
        filled in.

            >>> response = Request(connection, interface).call("Foo", bar="bar")
            >>> if response.response != 0:
            ...     print("some error occured")

        The DBus call itself is asynchronous (required for signals to work)
        but this method does not return until the Response is received, the
        Request is closed or an error occurs. If the Request is closed, the
        Response is None.

        If the "reply_handler" and "error_handler" keywords are present, those
        callbacks are called just like they would be as dbus.service.ProxyObject.
        """
        assert not self.used
        self.used = True

        # Make sure options exists and has the handle_token set
        try:
            options = kwargs["options"]
        except KeyError:
            options = dbus.Dictionary({}, signature="sv")

        if "handle_token" not in options:
            options["handle_token"] = self.handle_token

        # Anything that takes longer than 5s needs to fail
        self._mainloop = GLib.MainLoop()
        GLib.timeout_add(
            DBUS_TIMEOUT * 10 if is_valgrind() else DBUS_TIMEOUT, self._mainloop.quit
        )

        method = getattr(self.interface, methodname)
        assert method

        reply_handler = kwargs.pop("reply_handler", None)
        error_handler = kwargs.pop("error_handler", None)

        # Handle the normal method reply which returns is the Request object
        # path. We don't exit the mainloop here, we're waiting for either the
        # Response signal on the Request itself or the Close() handling
        def reply_cb(handle):
            try:
                logger.debug(f"Reply to {methodname} with {self.handle}")
                assert handle == self.handle

                if reply_handler:
                    reply_handler(handle)
            except Exception as e:
                self.error = e

        # Handle any exceptions during the actual method call (not the Request
        # handling itself). Can exit the mainloop if that happens
        def error_cb(error):
            try:
                logger.debug(f"Error after {methodname} with {error}")
                if error_handler:
                    error_handler(error)
                self.error = error
            except Exception as e:
                self.error = e
            finally:
                if self._mainloop:
                    self._mainloop.quit()

        # Method is invoked async, otherwise we can't mix and match signals
        # and other calls. It's still sync as seen by the caller in that we
        # have a mainloop that waits for us to finish though.
        method(
            *list(kwargs.values()),
            reply_handler=reply_cb,
            error_handler=error_cb,
        )

        self._mainloop.run()

        if self.error:
            raise self.error
        elif not self.closed and self.response is None:
            raise ResponseTimeout(f"Timed out waiting for response from {methodname}")

        return self.response


class Session(Closable):
    """
    Helper class for a Session created by a portal. This class takes care of
    subscribing to the `Closed` signals. A typical invocation is:

        >>> response = Request(connection, interface).call("CreateSession")
        >>> session = Session.from_response(response)
        # Now run the main loop and do other stuff
        # Check if the session was closed
        >>> if session.closed:
        ...    pass
        # or close the session explicitly
        >>> session.close()  # to close the session or
    """

    def __init__(self, bus: dbus.Bus, handle: str):
        assert handle
        super().__init__(bus, handle)

        self.handle = handle
        self.details = None
        # GLib makes assertions in callbacks impossible, so we wrap all
        # callbacks into a try: except and store the error on the request to
        # be raised later when we're back in the main context
        self.error = None
        self._closed_sig_received = False

        def cb_closed(details: ASV) -> None:
            try:
                logger.debug(f"Session.Closed received on {self.handle}")
                assert not self._closed_sig_received
                self._closed_sig_received = True
                self.details = details
                if self._mainloop:
                    self._mainloop.quit()
            except Exception as e:
                self.error = e

        proxy = bus.get_object("org.freedesktop.portal.Desktop", handle)
        self.session_interface = dbus.Interface(proxy, "org.freedesktop.portal.Session")
        self.session_interface.connect_to_signal("Closed", cb_closed)

    @property
    def closed(self):
        """
        Returns True if the session was closed by the backend
        """
        return self._closed_sig_received or super().closed

    @classmethod
    def from_response(cls, bus: dbus.Bus, response: Response) -> "Session":
        return cls(bus, response.results["session_handle"])


class GDBusIfaceSignal:
    """
    Helper class which represents a connected signal on a GDBusIface and can be
    used to disconnect from the signal.
    """

    def __init__(self, signal_id: int, proxy: Gio.DBusProxy):
        self.signal_id = signal_id
        self.proxy = proxy

    def disconnect(self):
        """
        Disconnects the signal
        """
        self.proxy.disconnect(self.signal_id)


class GDBusIface:
    """
    Helper class for calling dbus interfaces with complex arguments.
    Usually you want to use python-dbus on the dbus_con fixture with
    get_portal_iface , get_mock_iface or get_iface. This is convenient but
    might not be sufficient for complex arguments or for asynchronously calling
    a method.
    """

    def __init__(self, bus: str, obj: str, iface: str):
        """
        Creates a GDBusIface for a specific bus, object and interface on the
        session bus.
        """
        address = Gio.dbus_address_get_for_bus_sync(Gio.BusType.SESSION, None)
        session_bus = Gio.DBusConnection.new_for_address_sync(
            address,
            Gio.DBusConnectionFlags.AUTHENTICATION_CLIENT
            | Gio.DBusConnectionFlags.MESSAGE_BUS_CONNECTION,
            None,
            None,
        )
        assert session_bus
        self._proxy = Gio.DBusProxy.new_sync(
            session_bus,
            Gio.DBusProxyFlags.NONE,
            None,
            bus,
            obj,
            iface,
            None,
        )

    def _call(
        self, method_name: str, args_variant: GLib.Variant, fds: List[int] = []
    ) -> GLib.Variant:
        """
        Calls a method synchronously with the arguments passed in args_variant,
        passing the file descriptors specified in fds.
        Returns the result of the dbus call.
        """
        fdlist = Gio.UnixFDList.new()
        for fd in fds:
            fdlist.append(fd)

        return self._proxy.call_with_unix_fd_list_sync(
            method_name,
            args_variant,
            0,
            -1,
            fdlist,
            None,
        )

    def _call_async(
        self,
        method_name: str,
        args_variant: GLib.Variant,
        fds: List[int] = [],
        cb: Optional[Callable[[GLib.Variant], None]] = None,
    ) -> None:
        """
        Calls a method asynchronously with the arguments passed in args_variant,
        passing the file descriptors specified in fds.
        Invokes the callback cb when the call finished.
        """
        fdlist = Gio.UnixFDList.new()
        for fd in fds:
            fdlist.append(fd)

        def internal_cb(s, res, _):
            res = s.call_finish(res)
            if cb:
                cb(res)

        self._proxy.call_with_unix_fd_list(
            method_name,
            args_variant,
            0,
            -1,
            fdlist,
            None,
            internal_cb,
            None,
        )

    def connect_to_signal(
        self, name: str, cb: Callable[[GLib.Variant], None]
    ) -> GDBusIfaceSignal:
        """
        Connects to the dbus signal name to the callback cb. Returns an object
        representing the connection which can be used to disconnect it again.
        """

        def internal_cb(proxy, sender_name, signal_name, parameters):
            if signal_name != name:
                return
            cb(parameters)

        signal_id = self._proxy.connect("g-signal", internal_cb)
        return GDBusIfaceSignal(signal_id, self._proxy)


class FileAccessMode(Enum):
    READ_WRITE = "read-write"
    READ_ONLY = "read-only"
    HIDDEN = "hidden"


class ExecutableMock(object):
    def __init__(self, executable: str, access_mode: FileAccessMode):
        self.executable = executable
        self.access_mode = access_mode

    def create(self, path: Path):
        self.path = path / self.executable
        self.path.write_bytes(self.get_executable())
        self.path.chmod(0o755)

    def get_executable(self):
        return f"#!/usr/bin/env sh\necho {self.access_mode.value}".encode("utf8")


class SessionPersistenceMode(IntEnum):
    NONE = 0
    TRANSIENT = 1
    PERSISTENT = 2



# SPDX-License-Identifier: LGPL-2.1-or-later
# mypy: disable-error-code="misc"

from tests.templates.xdp_utils import init_logger

import dbus.service
import dbus

from dbusmock import mockobject

BUS_NAME = "org.freedesktop.GeoClue2"
MAIN_OBJ = "/org/freedesktop/GeoClue2/Manager"
MAIN_IFACE = "org.freedesktop.GeoClue2.Manager"
CLIENT_IFACE = "org.freedesktop.GeoClue2.Client"
LOCATION_IFACE = "org.freedesktop.GeoClue2.Location"
MOCK_IFACE = "org.freedesktop.GeoClue2.Mock"
SYSTEM_BUS = True
VERSION = 1

logger = init_logger(__name__)


class GeoClueClient(mockobject.DBusMockObject):
    def __init__(self, *args, **kwargs):
        super(GeoClueClient, self).__init__(*args, **kwargs)

        self.started = False
        self.location = 0
        self.data = {"Latitude": 0, "Longitude": 0, "Accuracy": 0}

    @dbus.service.method(
        CLIENT_IFACE,
        in_signature="",
        out_signature="",
    )
    def Start(self):
        logger.debug("Start()")
        self.started = True
        self.ChangeLocation(self.data)

    @dbus.service.method(
        CLIENT_IFACE,
        in_signature="",
        out_signature="",
    )
    def Stop(self):
        logger.debug("Stop()")
        self.started = False
        self.RemoveObject(f"/org/freedesktop/GeoClue2/Location/{self.location}")

    @dbus.service.method(
        MOCK_IFACE,
        in_signature="a{sv}",
        out_signature="",
    )
    def ChangeLocation(self, data):
        logger.debug(f"ChangeLocation({data})")

        self.data = data

        if not self.started:
            return

        old_path = "/"
        if self.location > 0:
            old_path = f"/org/freedesktop/GeoClue2/Location/{self.location}"
        self.location = self.location + 1
        new_path = f"/org/freedesktop/GeoClue2/Location/{self.location}"

        self.AddObject(
            new_path,
            LOCATION_IFACE,
            {
                "Latitude": data["Latitude"],
                "Longitude": data["Longitude"],
                "Accuracy": data["Accuracy"],
            },
            [],
        )

        if old_path != "/":
            self.RemoveObject(old_path)

        self.EmitSignal(
            CLIENT_IFACE,
            "LocationUpdated",
            "oo",
            [
                dbus.ObjectPath(old_path),
                dbus.ObjectPath(new_path),
            ],
        )


def load(mock, parameters={}):
    mock.AddMethods(
        MAIN_IFACE,
        [
            (
                "GetClient",
                "",
                "o",
                'ret = dbus.ObjectPath("/org/freedesktop/GeoClue2/Client/1")',
            ),
        ],
    )
    mock.AddObject(
        "/org/freedesktop/GeoClue2/Client/1",
        CLIENT_IFACE,
        {
            "DesktopId": "",
            "DistanceThreshold": 0,
            "TimeThreshold": 0,
            "RequestedAccuracyLevel": 0,
        },
        [],
        mock_class=GeoClueClient,
    )



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import pytest
import os
from pathlib import Path


ACCOUNT_DATA = {
    "id": "test",
    "name": "Test Name",
    "image": "FILLED OUT LATER",
}


@pytest.fixture
def required_templates():
    image = Path(os.environ["XDG_DATA_HOME"]) / "account-image.png"
    image.write_text("image contents")
    ACCOUNT_DATA["image"] = f"file://{image.absolute().as_posix()}"

    return {
        "account": {
            "results": ACCOUNT_DATA,
        },
    }


class TestAccount:
    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Account", 1)

    def test_basic1(self, xdg_document_portal, portals, dbus_con, xdp_app_info):
        app_id = xdp_app_info.app_id
        account_intf = xdp.get_portal_iface(dbus_con, "Account")
        mock_intf = xdp.get_mock_iface(dbus_con)

        reason = "reason"

        request = xdp.Request(dbus_con, account_intf)
        options = {
            "reason": reason,
        }
        response = request.call(
            "GetUserInformation",
            window="",
            options=options,
        )

        assert response
        assert response.response == 0
        assert response.results["id"] == ACCOUNT_DATA["id"]
        assert response.results["name"] == ACCOUNT_DATA["name"]
        assert response.results["image"]
        assert xdp.uri_same_file(ACCOUNT_DATA["image"], response.results["image"])

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("GetUserInformation")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[1] == app_id
        assert args[2] == ""  # window
        assert args[3]["reason"] == reason

    def test_reason(self, portals, dbus_con):
        account_intf = xdp.get_portal_iface(dbus_con, "Account")
        mock_intf = xdp.get_mock_iface(dbus_con)

        reason = """This reason is unreasonably long, it stretches over
                more than twohundredfiftysix characters, which is really quite
                long. Excessively so. The portal frontend will silently drop
                reasons of this magnitude. If you can't express your reasons
                concisely, you probably have no good reason in the first place
                and are just waffling around."""

        assert len(reason) - 1 > 256

        request = xdp.Request(dbus_con, account_intf)
        options = {
            "reason": reason,
        }
        response = request.call(
            "GetUserInformation",
            window="",
            options=options,
        )

        assert response
        assert response.response == 0

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("GetUserInformation")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert "reason" not in args[3]

    @pytest.mark.parametrize("template_params", ({"account": {"expect-close": True}},))
    def test_close(self, portals, dbus_con):
        account_intf = xdp.get_portal_iface(dbus_con, "Account")

        reason = "reason"

        request = xdp.Request(dbus_con, account_intf)
        request.schedule_close(1000)
        options = {
            "reason": reason,
        }
        request.call(
            "GetUserInformation",
            window="",
            options=options,
        )

        # Only true if the impl.Request was closed too
        assert request.closed



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import pytest
import os
import gi
import subprocess
import re
from typing import assert_never

gi.require_version("UMockdev", "1.0")
from gi.repository import GLib, UMockdev  # noqa E402


@pytest.fixture
def required_templates():
    return {"usb": {}}


@pytest.fixture
def usb_queries() -> str | None:
    return None


# Supported on host and flatpak
# Host apps get access to all usb devices. The query has no effect.
@pytest.fixture(params=[xdp.AppInfoKind.HOST, xdp.AppInfoKind.FLATPAK])
def xdp_app_info(request, usb_queries) -> xdp.AppInfo:
    app_info_kind = request.param
    app_id = "org.example.Test"

    if app_info_kind == xdp.AppInfoKind.HOST:
        return xdp.AppInfo.new_host(
            app_id=app_id,
        )

    if app_info_kind == xdp.AppInfoKind.FLATPAK:
        return xdp.AppInfo.new_flatpak(
            app_id=app_id,
            usb_queries=usb_queries,
        )

    assert_never(app_info_kind)


@pytest.fixture
def umockdev():
    return UMockdev.Testbed.new()


def umockdev_has_working_remove():
    # umockdev only generates remove events since version 0.18.4
    # https://github.com/martinpitt/umockdev/releases/tag/0.18.4
    required = (0, 18, 4)

    result = subprocess.run(["umockdev-run", "--version"], stdout=subprocess.PIPE)
    if result.returncode != 0:
        return False
    match = re.match(r"^(\d+)\.(\d+)\.(\d+)", result.stdout.decode("UTF-8").strip())
    if not match:
        return False
    version = tuple(map(int, match.groups()))
    return version >= required


class TestUsb:
    _num_devices = 0

    def generate_device(
        self, testbed, vendor, vendor_name, product, product_name, serial
    ):
        n = self._num_devices
        self._num_devices += 1

        testbed.add_from_string(f"""P: /devices/usb{n}
N: bus/usb/001/{n:03d}
E: BUSNUM=001
E: DEVNUM={n:03d}
E: DEVNAME=/dev/bus/usb/001/{n:03d}
E: DEVTYPE=usb_device
E: DRIVER=usb
E: ID_BUS=usb
E: ID_MODEL={product_name}
E: ID_MODEL_ID={product}
E: ID_REVISION=0002
E: ID_SERIAL={vendor_name}_{product_name}_{serial}
E: ID_SERIAL_SHORT={serial}
E: ID_VENDOR={vendor_name}
E: ID_VENDOR_ID={vendor}
E: SUBSYSTEM=usb
A: idProduct={product}
A: idVendor={vendor}
""")

        return f"/sys/devices/usb{n}"

    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Usb", 1)

    def test_create_close_session(self, portals, dbus_con):
        usb_intf = xdp.get_portal_iface(dbus_con, "Usb")

        session = xdp.Session(
            dbus_con,
            usb_intf.CreateSession({"session_handle_token": "session_token0"}),
        )
        session.close()

    def test_empty_initial_devices(self, portals, dbus_con):
        usb_intf = xdp.get_portal_iface(dbus_con, "Usb")

        xdp.Session(
            dbus_con,
            usb_intf.CreateSession({"session_handle_token": "session_token0"}),
        )

        device_events_signal_received = False

        def cb_device_events(session_handle, events):
            nonlocal device_events_signal_received
            device_events_signal_received = True

        usb_intf.connect_to_signal("DeviceEvents", cb_device_events)
        xdp.wait(300)
        assert not device_events_signal_received

    @pytest.mark.parametrize("usb_queries", ["vnd:04a9", None])
    def test_initial_devices(
        self, portals, dbus_con, xdp_app_info, usb_queries, umockdev
    ):
        usb_intf = xdp.get_portal_iface(dbus_con, "Usb")

        self.generate_device(
            umockdev,
            "04a9",
            "Canon_Inc.",
            "31c0",
            "Canon_Digital_Camera",
            "C767F1C714174C309255F70E4A7B2EE2",
        )

        # TODO: to make this more robust, we should find a way to wait for
        # the portal to pick up the device
        xdp.wait(300)

        session = xdp.Session(
            dbus_con,
            usb_intf.CreateSession({"session_handle_token": "session_token0"}),
        )

        device_events_signal_received = False
        devices_received = 0

        def cb_device_events(session_handle, events):
            nonlocal device_events_signal_received
            nonlocal devices_received
            assert session.handle == session_handle

            for action, id, device in events:
                assert action == "add"
                devices_received += 1

            device_events_signal_received = True

        usb_intf.connect_to_signal("DeviceEvents", cb_device_events)

        # If the app is not running on the host, make sure an empty usb_query
        # results in no devices being available
        if usb_queries is None and xdp_app_info.kind != xdp.AppInfoKind.HOST:
            xdp.wait(300)
            assert not device_events_signal_received
            assert devices_received == 0
        else:
            xdp.wait_for(lambda: device_events_signal_received)
            assert devices_received == 1

    @pytest.mark.parametrize("usb_queries", ["vnd:04a9", None])
    def test_device_add(self, portals, xdp_app_info, dbus_con, usb_queries, umockdev):
        usb_intf = xdp.get_portal_iface(dbus_con, "Usb")

        session = xdp.Session(
            dbus_con,
            usb_intf.CreateSession({"session_handle_token": "session_token0"}),
        )

        device_events_signal_received = False
        devices_received = 0
        device = None

        def cb_device_events(session_handle, events):
            nonlocal device_events_signal_received
            nonlocal devices_received
            nonlocal device
            assert session.handle == session_handle

            for action, _, dev in events:
                assert action == "add"
                device = dev
                devices_received += 1

            device_events_signal_received = True

        usb_intf.connect_to_signal("DeviceEvents", cb_device_events)
        xdp.wait(300)
        assert not device_events_signal_received

        self.generate_device(
            umockdev,
            "04a9",
            "Canon_Inc.",
            "31c0",
            "Canon_Digital_Camera",
            "C767F1C714174C309255F70E4A7B2EE2",
        )

        if usb_queries is None and xdp_app_info.kind != xdp.AppInfoKind.HOST:
            xdp.wait(300)
            assert not device_events_signal_received
            assert devices_received == 0
        else:
            xdp.wait_for(lambda: device_events_signal_received)
            assert devices_received == 1

            assert device
            assert device["readable"]
            assert device["writable"]
            assert device["device-file"] == "/dev/bus/usb/001/000"
            assert device["properties"]["ID_VENDOR_ID"] == "04a9"
            assert device["properties"]["ID_MODEL_ID"] == "31c0"
            assert (
                device["properties"]["ID_SERIAL"]
                == "Canon_Inc._Canon_Digital_Camera_C767F1C714174C309255F70E4A7B2EE2"
            )

    @pytest.mark.parametrize("usb_queries", ["vnd:04a9", None])
    @pytest.mark.skipif(
        not umockdev_has_working_remove(), reason="UMockdev version 0.18.4 required"
    )
    def test_device_remove(
        self, portals, dbus_con, xdp_app_info, usb_queries, umockdev
    ):
        usb_intf = xdp.get_portal_iface(dbus_con, "Usb")

        dev_path = self.generate_device(
            umockdev,
            "04a9",
            "Canon_Inc.",
            "31c0",
            "Canon_Digital_Camera",
            "C767F1C714174C309255F70E4A7B2EE2",
        )

        session = xdp.Session(
            dbus_con,
            usb_intf.CreateSession({"session_handle_token": "session_token0"}),
        )

        device_events_signal_count = 0
        devices_received = 0
        devices_removed = 0

        def cb_device_events(session_handle, events):
            nonlocal device_events_signal_count
            nonlocal devices_received
            nonlocal devices_removed

            assert session.handle == session_handle

            for action, id, device in events:
                if action == "add":
                    devices_received += 1
                elif action == "remove":
                    devices_removed += 1
                else:
                    assert False

            device_events_signal_count += 1

        usb_intf.connect_to_signal("DeviceEvents", cb_device_events)

        if usb_queries is None and xdp_app_info.kind != xdp.AppInfoKind.HOST:
            xdp.wait(300)
            assert device_events_signal_count == 0
            assert devices_received == 0
            assert devices_removed == 0
        else:
            xdp.wait_for(lambda: device_events_signal_count == 1)
            assert devices_received == 1
            assert devices_removed == 0

        umockdev.remove_device(dev_path)

        if usb_queries is None and xdp_app_info.kind != xdp.AppInfoKind.HOST:
            xdp.wait(300)
            assert device_events_signal_count == 0
            assert devices_received == 0
            assert devices_removed == 0
        else:
            xdp.wait_for(lambda: device_events_signal_count == 2)
            assert devices_received == 1
            assert devices_removed == 1

    @pytest.mark.parametrize("usb_queries", ["vnd:04a9;vnd:04aa"])
    @pytest.mark.parametrize(
        "template_params", [{"usb": {"filters": {"vendor": "04a9"}}}]
    )
    def test_acquire(self, portals, dbus_con, umockdev):
        usb_intf = xdp.get_portal_iface(dbus_con, "Usb")

        self.generate_device(
            umockdev,
            "04a9",
            "Canon_Inc.",
            "31c0",
            "Canon_Digital_Camera",
            "C767F1C714174C309255F70E4A7B2EE2",
        )

        self.generate_device(
            umockdev,
            "04aa",
            "Someone Else.",
            "31c0",
            "SomeProduct",
            "00001",
        )

        possible_vendors = ["04a9", "04aa"]

        devices = usb_intf.EnumerateDevices({})
        assert len(devices) == 2
        (id1, dev_info1) = devices[0]
        assert id1
        assert dev_info1
        vendor_id = dev_info1["properties"]["ID_VENDOR_ID"]
        assert vendor_id in possible_vendors
        possible_vendors.remove(vendor_id)
        (id2, dev_info2) = devices[1]
        assert id2
        assert dev_info2
        vendor_id = dev_info2["properties"]["ID_VENDOR_ID"]
        assert vendor_id in possible_vendors
        possible_vendors.remove(vendor_id)

        request = xdp.Request(dbus_con, usb_intf)
        response = request.call(
            "AcquireDevices",
            parent_window="",
            devices=[
                (id1, {"writable": True}),
                (id2, {"writable": True}),
            ],
            options={},
        )
        assert response
        assert response.response == 0

        (results, finished) = usb_intf.FinishAcquireDevices(request.handle, {})
        assert finished
        assert len(results) == 1
        (res_id, device) = results[0]
        assert res_id == id1 or res_id == id2
        assert device["success"]
        fd = device["fd"].take()
        assert fd > 0
        with os.fdopen(fd, "r") as f:
            assert f
        assert "error" not in device

        usb_intf.ReleaseDevices([res_id], {})

    @pytest.mark.parametrize("usb_queries", ["vnd:0001"])
    @pytest.mark.parametrize(
        "expected,template_params",
        [
            (1, {"usb": {"filters": {"model": "0000"}}}),
            (1, {"usb": {"filters": {"model": "0001"}}}),
            (0, {"usb": {"filters": {"model": "0002"}}}),
            (2, {"usb": {"filters": {"vendor": "0001"}}}),
            (0, {"usb": {"filters": {"vendor": "0002"}}}),
            (1, {"usb": {"filters": {"vendor": "0001", "model": "0000"}}}),
            (0, {"usb": {"filters": {"vendor": "0002", "model": "0000"}}}),
        ],
    )
    def test_queries(self, portals, dbus_con, expected, usb_queries, umockdev):
        usb_intf = xdp.get_portal_iface(dbus_con, "Usb")

        for i in range(2):
            self.generate_device(
                umockdev,
                "0001",
                "example_org",
                f"000{i}",
                f"model{i}",
                "0001",
            )

        devices = usb_intf.EnumerateDevices({})
        assert len(devices) == 2
        acquire_devices = [(id, {"writable": True}) for (id, _) in devices]

        request = xdp.Request(dbus_con, usb_intf)
        response = request.call(
            "AcquireDevices",
            parent_window="",
            devices=acquire_devices,
            options={},
        )
        assert response
        assert response.response == 0

        (results, finished) = usb_intf.FinishAcquireDevices(request.handle, {})
        assert finished
        assert len(results) == expected



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import os
import pytest
import tempfile
from pathlib import Path
from gi.repository import GLib


class TestTrash:
    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "Trash", 1)

    def test_trash_file_fails(self, portals, dbus_con):
        trash_intf = xdp.get_portal_iface(dbus_con, "Trash")
        try:
            with open("/proc/cmdline") as fd:
                result = trash_intf.TrashFile(fd.fileno())
        except PermissionError as e:
            pytest.skip(f"Couldn't open file /proc/cmdline: {e}")

        assert result == 0

    def test_trash_file(self, portals, dbus_con):
        trash_intf = xdp.get_portal_iface(dbus_con, "Trash")

        fd, name = tempfile.mkstemp(prefix="trash_portal_mock_", dir=Path.home())
        result = trash_intf.TrashFile(fd)

        assert result == 1
        assert not Path(name).exists()

        info_dir = Path(os.environ["XDG_DATA_HOME"]) / "Trash/info"
        assert info_dir.exists()

        files_dir = Path(os.environ["XDG_DATA_HOME"]) / "Trash/files"
        assert files_dir.exists()

        trashed_files = info_dir.iterdir()
        trashed_file = next(trashed_files)

        keyfile = GLib.KeyFile.new()
        content = trashed_file.read_text()
        assert keyfile.load_from_data(
            content,
            len(content),
            GLib.KeyFileFlags.NONE,
        )
        assert keyfile.get_string("Trash Info", "Path") == name
        assert (files_dir / trashed_file.stem).exists()

        with pytest.raises(StopIteration):
            next(trashed_files)

    def test_trash_folder(self, portals, dbus_con):
        trash_intf = xdp.get_portal_iface(dbus_con, "Trash")

        folder = Path(os.environ["HOME"]) / "folder-to-trash"
        file_in_folder = folder / "foo" / "bar" / "file"
        file_in_folder.parent.mkdir(parents=True)
        file_in_folder.write_text("foobar")

        fd = os.open(folder, os.O_RDONLY | os.O_CLOEXEC)
        try:
            result = trash_intf.TrashFile(fd)
        finally:
            os.close(fd)

        assert result == 1
        assert not Path(folder).exists()

        info_dir = Path(os.environ["XDG_DATA_HOME"]) / "Trash/info"
        assert info_dir.exists()

        files_dir = Path(os.environ["XDG_DATA_HOME"]) / "Trash/files"
        assert files_dir.exists()

        trashed_files = info_dir.iterdir()
        trashed_file = next(trashed_files)

        keyfile = GLib.KeyFile.new()
        content = trashed_file.read_text()
        assert keyfile.load_from_data(
            content,
            len(content),
            GLib.KeyFileFlags.NONE,
        )
        assert keyfile.get_string("Trash Info", "Path") == folder.as_posix()

        trashed_folder = files_dir / trashed_file.stem
        assert trashed_folder.exists()
        trashed_file = trashed_folder / "foo" / "bar" / "file"
        assert trashed_file.exists()

        with pytest.raises(StopIteration):
            next(trashed_files)



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import dbus
import pytest
import socket
from gi.repository import GLib
from itertools import count
from typing import Any


counter = count()


def default_zones():
    return [(1024, 768, 0, 0), (640, 480, 1024, 0)]


@pytest.fixture
def required_templates():
    return {"inputcapture": {}}


@pytest.fixture
def zones():
    return default_zones()


class InputcaptureSession:
    def __init__(self, dbus_con):
        self.dbus_con = dbus_con
        self.type = None
        self.session = None
        self.session_handle = None
        self.current_zone_set = []

    def create_legacy(self, capabilities=0x7):
        """
        Call CreateSession for the given capabilities and return the
        (response, results) tuple.
        """

        assert self.type is None
        self.type = "legacy"

        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")
        mock_intf = xdp.get_mock_iface(self.dbus_con)

        request = xdp.Request(self.dbus_con, inputcapture_intf)

        capabilities = dbus.UInt32(capabilities, variant_level=1)
        session_handle_token = dbus.String(f"session{next(counter)}", variant_level=1)

        options = dbus.Dictionary(
            {
                "capabilities": capabilities,
                "session_handle_token": session_handle_token,
            },
            signature="sv",
        )

        response = request.call("CreateSession", parent_window="", options=options)

        assert response
        assert response.response == 0
        assert "session_handle" in response.results
        assert "capabilities" in response.results
        caps = response.results["capabilities"]
        # Returned capabilities must be a subset of the requested ones
        assert caps & ~capabilities == 0

        self.session = xdp.Session(self.dbus_con, response.results["session_handle"])
        self.session_handle = self.session.handle

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Start")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[3] == ""  # parent window
        assert args[4]["capabilities"] == capabilities

        return response

    def create(self):
        """
        Call CreateSession for the given capabilities and return the
        (response, results) tuple.
        """

        assert self.type is None
        self.type = "modern"

        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")

        session_handle_token = dbus.String(f"session{next(counter)}", variant_level=1)

        result = inputcapture_intf.CreateSession2(
            {"session_handle_token": session_handle_token}
        )
        self.session = xdp.Session(self.dbus_con, result["session_handle"])
        self.session_handle = self.session.handle

    def start(
        self,
        capabilities=0x7,
        persist_mode: xdp.SessionPersistenceMode | None = None,
        restore_token: str | None = None,
        validate_token_data: bool = True,
    ):
        assert self.type == "modern"

        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")
        mock_intf = xdp.get_mock_iface(self.dbus_con)

        capabilities = dbus.UInt32(capabilities, variant_level=1)
        request = xdp.Request(self.dbus_con, inputcapture_intf)
        options = {
            "capabilities": capabilities,
        }
        if persist_mode is not None:
            options["persist_mode"] = dbus.UInt32(persist_mode, variant_level=1)
        if restore_token is not None:
            options["restore_token"] = restore_token
        response = request.call(
            "Start",
            session_handle=self.session_handle,
            parent_window="",
            options=options,
        )

        assert response
        assert response.response == 0
        assert "capabilities" in response.results
        caps = response.results["capabilities"]
        # Returned capabilities must be a subset of the requested ones
        assert caps & ~capabilities == 0

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Start")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[3] == ""  # parent window
        options = args[4]
        assert options["capabilities"] == capabilities
        if persist_mode is not None:
            assert options["persist_mode"] == dbus.UInt32(persist_mode, variant_level=1)
        if restore_token is not None and validate_token_data:
            restore_data = options["restore_data"]
            assert restore_data is not None
            # Vendor-name and version are hardcoded in the template
            assert restore_data[0] == "GNOME"
            assert restore_data[1] == dbus.UInt32(7)

        return response

    def get_zones(self):
        """
        Call GetZones and return the (response, results) tuple.
        """
        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")
        mock_intf = xdp.get_mock_iface(self.dbus_con)

        request = xdp.Request(self.dbus_con, inputcapture_intf)
        options: Any = {}
        response = request.call(
            "GetZones", session_handle=self.session_handle, options=options
        )

        assert response
        assert response.response == 0
        assert "zones" in response.results
        assert "zone_set" in response.results

        self.current_zone_set = response.results["zone_set"]

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("GetZones")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[0] == request.handle
        assert args[1] == self.session_handle

        return response

    def set_pointer_barriers(self, barriers):
        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")
        mock_intf = xdp.get_mock_iface(self.dbus_con)

        request = xdp.Request(self.dbus_con, inputcapture_intf)
        options: Any = {}
        response = request.call(
            "SetPointerBarriers",
            session_handle=self.session_handle,
            options=options,
            barriers=barriers,
            zone_set=self.current_zone_set,
        )
        assert response
        assert response.response == 0
        assert "failed_barriers" in response.results

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("SetPointerBarriers")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[0] == request.handle
        assert args[1] == self.session_handle
        assert args[4] == barriers
        assert args[5] == self.current_zone_set

        return response

    def connect_to_eis(self):
        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")
        mock_intf = xdp.get_mock_iface(self.dbus_con)

        fd = inputcapture_intf.ConnectToEIS(
            self.session_handle, dbus.Dictionary({}, signature="sv")
        )

        # Our dbusmock template sends HELLO
        eis_socket = socket.fromfd(fd.take(), socket.AF_UNIX, socket.SOCK_STREAM)
        hello = eis_socket.recv(10)
        assert hello == b"HELLO"

        method_calls = mock_intf.GetMethodCalls("ConnectToEIS")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[0] == self.session_handle

        return eis_socket

    def enable(self):
        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")
        mock_intf = xdp.get_mock_iface(self.dbus_con)

        inputcapture_intf.Enable(
            self.session_handle, dbus.Dictionary({}, signature="sv")
        )

        method_calls = mock_intf.GetMethodCalls("Enable")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[0] == self.session_handle

    def disable(self):
        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")
        mock_intf = xdp.get_mock_iface(self.dbus_con)

        inputcapture_intf.Disable(
            self.session_handle, dbus.Dictionary({}, signature="sv")
        )

        method_calls = mock_intf.GetMethodCalls("Disable")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[0] == self.session_handle

    def release(self, activation_id: int, cursor_position=None):
        inputcapture_intf = xdp.get_portal_iface(self.dbus_con, "InputCapture")
        mock_intf = xdp.get_mock_iface(self.dbus_con)

        options = {"activation_id": dbus.UInt32(activation_id)}
        if cursor_position:
            options["cursor_position"] = dbus.Struct(
                list(cursor_position), signature="dd", variant_level=1
            )

        inputcapture_intf.Release(
            self.session_handle, dbus.Dictionary(options, signature="sv")
        )

        method_calls = mock_intf.GetMethodCalls("Release")
        assert len(method_calls) > 0
        _, args = method_calls[-1]
        assert args[0] == self.session_handle
        assert "activation_id" in args[2]
        aid = args[2]["activation_id"]
        assert aid == activation_id
        if cursor_position:
            assert "cursor_position" in args[2]
            pos = args[2]["cursor_position"]
            assert pos == cursor_position


class TestInputCapture:
    def test_version(self, portals, dbus_con):
        xdp.check_version(dbus_con, "InputCapture", 2)

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "inputcapture": {
                    "supported_capabilities": 0b101,  # KEYBOARD, POINTER, TOUCH
                },
            },
        ),
    )
    def test_supported_capabilities(self, portals, dbus_con):
        properties_intf = xdp.get_iface(dbus_con, "org.freedesktop.DBus.Properties")

        caps = properties_intf.Get(
            "org.freedesktop.portal.InputCapture", "SupportedCapabilities"
        )
        assert caps == 0b101

    def test_create_session(self, portals, dbus_con):
        session = InputcaptureSession(dbus_con)
        session.create()  # KEYBOARD
        session.start(capabilities=0b1)  # KEYBOARD
        session = InputcaptureSession(dbus_con)
        session.create_legacy(capabilities=0b1)  # KEYBOARD

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "inputcapture": {
                    "capabilities": 0b110,  # TOUCH, POINTER
                    "supported_capabilities": 0b111,  # TOUCH, POINTER, KEYBOARD
                },
            },
        ),
    )
    def test_create_session_limited_caps(self, portals, dbus_con):
        # Request more caps than are supported
        session = InputcaptureSession(dbus_con)
        session.create()
        response, results = session.start(capabilities=0b111)
        caps = results["capabilities"]
        # Returned capabilities must the ones we set up in the params
        assert caps == 0b110

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "inputcapture": {
                    "default-zone": dbus.Array(
                        [dbus.Struct(z, signature="uuii") for z in default_zones()],
                        signature="(uuii)",
                        variant_level=1,
                    )
                },
            },
        ),
    )
    def test_get_zones(self, portals, dbus_con, zones):
        mock_intf = xdp.get_mock_iface(dbus_con)

        session = InputcaptureSession(dbus_con)
        session.create()
        response, results = session.start()
        response, results = session.get_zones()
        for z1, z2 in zip(results["zones"], zones):
            assert z1 == z2

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Start")
        assert len(method_calls) == 1
        method_calls = mock_intf.GetMethodCalls("GetZones")
        assert len(method_calls) == 1

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "inputcapture": {
                    "default-zone": dbus.Array(
                        [dbus.Struct(z, signature="uuii") for z in default_zones()],
                        signature="(uuii)",
                        variant_level=1,
                    )
                },
            },
        ),
    )
    def test_set_pointer_barriers(self, portals, dbus_con, zones):
        mock_intf = xdp.get_mock_iface(dbus_con)

        session = InputcaptureSession(dbus_con)
        session.create()
        response, results = session.start()
        response, results = session.get_zones()

        barriers = [
            {
                "barrier_id": dbus.UInt32(10, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 0, 768], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(11, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1024, 0], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(12, variant_level=1),
                "position": dbus.Struct(
                    [1024, 0, 1024, 768], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(13, variant_level=1),
                "position": dbus.Struct(
                    [0, 768, 1024, 768], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(14, variant_level=1),
                "position": dbus.Struct(
                    [100, 768, 500, 768], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(15, variant_level=1),
                "position": dbus.Struct(
                    [1024, 0, 1024, 480], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(16, variant_level=1),
                "position": dbus.Struct(
                    [1024 + 640, 0, 1024 + 640, 480], signature="iiii", variant_level=1
                ),
            },
            # invalid ones
            {
                "barrier_id": dbus.UInt32(20, variant_level=1),
                "position": dbus.Struct(
                    [0, 1, 3, 4], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(21, variant_level=1),
                "position": dbus.Struct(
                    [0, 1, 1024, 1], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(22, variant_level=1),
                "position": dbus.Struct(
                    [1, 0, 1, 768], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(23, variant_level=1),
                "position": dbus.Struct(
                    [1023, 0, 1023, 768], signature="iiii", variant_level=1
                ),
            },
            {
                "barrier_id": dbus.UInt32(24, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1050, 0], signature="iiii", variant_level=1
                ),
            },
        ]
        response, results = session.set_pointer_barriers(barriers=barriers)
        failed_barriers = results["failed_barriers"]
        assert all([id >= 20 for id in failed_barriers])

        for id in [b["barrier_id"] for b in barriers if b["barrier_id"] >= 20]:
            assert id in failed_barriers

        # Check the impl portal was called with the right args
        method_calls = mock_intf.GetMethodCalls("Start")
        assert len(method_calls) == 1
        method_calls = mock_intf.GetMethodCalls("GetZones")
        assert len(method_calls) == 1
        method_calls = mock_intf.GetMethodCalls("SetPointerBarriers")
        assert len(method_calls) == 1
        _, args = method_calls.pop(0)
        assert args[4] == barriers
        assert args[5] == session.current_zone_set

    def test_connect_to_eis(self, portals, dbus_con):
        session = InputcaptureSession(dbus_con)
        session.create()
        session.start()
        session.get_zones()

        # The default zone is 1920x1080
        barriers = [
            {
                "barrier_id": dbus.UInt32(10, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1920, 0], signature="iiii", variant_level=1
                ),
            },
        ]
        session.set_pointer_barriers(barriers)

        session.connect_to_eis()

    def test_enable_disable(self, portals, dbus_con):
        mock_intf = xdp.get_mock_iface(dbus_con)

        session = InputcaptureSession(dbus_con)
        session.create()
        session.start()
        session.get_zones()

        # The default zone is 1920x1080
        barriers = [
            {
                "barrier_id": dbus.UInt32(10, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1920, 0], signature="iiii", variant_level=1
                ),
            },
        ]
        session.set_pointer_barriers(barriers)
        session.connect_to_eis()

        # Disable before enable should be a noop
        session.disable()

        method_calls = mock_intf.GetMethodCalls("Disable")
        assert len(method_calls) == 1

        session.enable()
        method_calls = mock_intf.GetMethodCalls("Enable")
        assert len(method_calls) == 1

        session.disable()
        method_calls = mock_intf.GetMethodCalls("Disable")
        assert len(method_calls) == 2

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "inputcapture": {
                    "disable-delay": 200,
                },
            },
        ),
    )
    def test_disable_signal(self, portals, dbus_con):
        inputcapture_intf = xdp.get_portal_iface(dbus_con, "InputCapture")

        session = InputcaptureSession(dbus_con)
        session.create()
        session.start()
        session.get_zones()
        # The default zone is 1920x1080
        barriers = [
            {
                "barrier_id": dbus.UInt32(10, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1920, 0], signature="iiii", variant_level=1
                ),
            },
        ]
        session.set_pointer_barriers(barriers)
        session.connect_to_eis()

        disabled_signal_received = False

        def cb_disabled(session_handle, options):
            nonlocal disabled_signal_received
            disabled_signal_received = True
            assert session_handle == session_handle

        inputcapture_intf.connect_to_signal("Disabled", cb_disabled)
        session.enable()
        xdp.wait_for(lambda: disabled_signal_received)

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "inputcapture": {
                    "activated-delay": 200,
                    "deactivated-delay": 300,
                },
            },
        ),
    )
    def test_activated_signal(self, portals, dbus_con):
        inputcapture_intf = xdp.get_portal_iface(dbus_con, "InputCapture")

        session = InputcaptureSession(dbus_con)
        session.create()
        session.start()
        session.get_zones()
        # The default zone is 1920x1080
        barriers = [
            {
                "barrier_id": dbus.UInt32(10, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1920, 0], signature="iiii", variant_level=1
                ),
            },
        ]
        session.set_pointer_barriers(barriers)
        session.connect_to_eis()

        disabled_signal_received = False
        activated_signal_received = False
        deactivated_signal_received = False

        def cb_disabled(session_handle, options):
            nonlocal disabled_signal_received
            disabled_signal_received = True

        def cb_activated(session_handle, options):
            nonlocal activated_signal_received
            activated_signal_received = True
            assert session_handle == session_handle
            assert "activation_id" in options
            assert "barrier_id" in options
            assert options["barrier_id"] == 10  # template uses first barrier
            assert "cursor_position" in options
            assert options["cursor_position"] == (
                10.0,
                20.0,
            )  # template uses x+10, y+20 of first barrier

        def cb_deactivated(session_handle, options):
            nonlocal deactivated_signal_received
            deactivated_signal_received = True
            assert session_handle == session_handle
            assert "activation_id" in options
            assert "cursor_position" in options
            assert options["cursor_position"] == (
                10.0,
                20.0,
            )  # template uses x+10, y+20 of first barrier

        inputcapture_intf.connect_to_signal("Activated", cb_activated)
        inputcapture_intf.connect_to_signal("Deactivated", cb_deactivated)
        inputcapture_intf.connect_to_signal("Disabled", cb_disabled)

        session.enable()

        xdp.wait_for(lambda: activated_signal_received and deactivated_signal_received)
        assert not disabled_signal_received

        # Disabling should not trigger the signal
        session.disable()
        assert not disabled_signal_received

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "inputcapture": {
                    "zones-changed-delay": 200,
                },
            },
        ),
    )
    def test_zones_changed_signal(self, portals, dbus_con):
        inputcapture_intf = xdp.get_portal_iface(dbus_con, "InputCapture")

        session = InputcaptureSession(dbus_con)
        session.create()
        session.start()
        session.get_zones()
        # The default zone is 1920x1080
        barriers = [
            {
                "barrier_id": dbus.UInt32(10, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1920, 0], signature="iiii", variant_level=1
                ),
            },
        ]
        session.set_pointer_barriers(barriers)
        session.connect_to_eis()

        zones_changed_signal_received = False

        def cb_zones_changed(session_handle, options):
            nonlocal zones_changed_signal_received
            zones_changed_signal_received = True
            assert session_handle == session_handle

        inputcapture_intf.connect_to_signal("ZonesChanged", cb_zones_changed)
        session.enable()
        xdp.wait_for(lambda: zones_changed_signal_received)

    @pytest.mark.parametrize(
        "template_params",
        (
            {
                "inputcapture": {
                    "activated-delay": 200,
                    "deactivated-delay": 1000,
                    "disabled-delay": 1200,
                },
            },
        ),
    )
    def test_release(self, portals, dbus_con):
        inputcapture_intf = xdp.get_portal_iface(dbus_con, "InputCapture")

        session = InputcaptureSession(dbus_con)
        session.create()
        session.start()
        session.get_zones()
        # The default zone is 1920x1080
        barriers = [
            {
                "barrier_id": dbus.UInt32(10, variant_level=1),
                "position": dbus.Struct(
                    [0, 0, 1920, 0], signature="iiii", variant_level=1
                ),
            },
        ]
        session.set_pointer_barriers(barriers)
        session.connect_to_eis()

        disabled_signal_received = False
        activated_signal_received = False
        deactivated_signal_received = False
        activation_id = None

        def cb_disabled(session_handle, options):
            nonlocal disabled_signal_received
            disabled_signal_received = True

        def cb_activated(session_handle, options):
            nonlocal activated_signal_received, activation_id
            activated_signal_received = True
            activation_id = options["activation_id"]

        def cb_deactivated(session_handle, options):
            nonlocal deactivated_signal_received
            deactivated_signal_received = True

        inputcapture_intf.connect_to_signal("Disabled", cb_disabled)
        inputcapture_intf.connect_to_signal("Activated", cb_activated)
        inputcapture_intf.connect_to_signal("Deactivated", cb_deactivated)

        session.enable()

        xdp.wait_for(lambda: activated_signal_received)
        assert activation_id is not None
        assert not deactivated_signal_received
        assert not disabled_signal_received

        session.release(cursor_position=(10.0, 50.0), activation_id=activation_id)

        # XDP should filter any signals the implementation may
        # send after Release().

        mainloop = GLib.MainLoop()
        GLib.timeout_add(1000, mainloop.quit)
        mainloop.run()

        # Release() implies deactivated
        assert not deactivated_signal_received
        assert not disabled_signal_received

    @pytest.mark.parametrize(
        "mode",
        [xdp.SessionPersistenceMode.TRANSIENT, xdp.SessionPersistenceMode.PERSISTENT],
    )
    def test_persistence(self, portals, dbus_con, mode):
        def create_persistent_session(
            persist_mode: xdp.SessionPersistenceMode,
            restore_token: str | None,
            validate_token_data: bool = True,
        ):
            ic_session = InputcaptureSession(dbus_con)
            ic_session.create()
            response = ic_session.start(
                restore_token=restore_token,
                persist_mode=persist_mode,
                validate_token_data=validate_token_data,
            )
            restore_token = response.results.get("restore_token")
            session = xdp.Session.from_response(dbus_con, response)
            session.close()
            return restore_token

        token1 = create_persistent_session(persist_mode=mode, restore_token=None)
        assert token1 is not None

        # Current implementation keeps the restore token for the same data
        token2 = create_persistent_session(persist_mode=mode, restore_token=token1)
        assert token2 is not None
        assert token2 == token1

        # A new session with different token data
        token3 = create_persistent_session(persist_mode=mode, restore_token=None)
        assert token3 is not None
        assert token3 != token1

        # First session again
        token4 = create_persistent_session(persist_mode=mode, restore_token=token1)
        assert token4 is not None
        assert token4 == token1

        # End the first session
        token5 = create_persistent_session(
            persist_mode=xdp.SessionPersistenceMode.NONE, restore_token=token1
        )
        assert token5 is None

        # token1 is invalid now
        token6 = create_persistent_session(
            persist_mode=xdp.SessionPersistenceMode.PERSISTENT,
            restore_token=token1,
            validate_token_data=False,
        )
        assert token6 is not None
        assert token6 != token1

        # but token3 is still there
        token7 = create_persistent_session(persist_mode=mode, restore_token=token3)
        assert token7 is not None
        assert token7 == token3



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import tests.xdp_utils as xdp

import pytest
import errno
import os
import random
import stat
import sys
import multiprocessing as mp
import traceback
from collections import defaultdict
from gi.repository import Gio, GLib


@pytest.fixture
def xdp_app_info() -> xdp.AppInfo:
    return xdp.AppInfo.new_host(
        app_id="",
    )


def filename_to_ay(filename):
    return list(filename.encode("utf-8")) + [0]


app_prefix = "org.test."
dir_prefix = "dir"
ensure_no_remaining = True

running_count: defaultdict[str, int] = defaultdict(int)


def log(str):
    print(str, file=sys.stderr)


def logv(str):
    log(str)


def get_a_count(counter: str):
    global running_count
    running_count[counter] += 1
    return running_count[counter]


def setFileContent(path, content):
    with open(path, "w") as f:
        f.write(content)


def appendFileContent(path, content):
    with open(path, "a") as f:
        f.write(content)


def readFdContent(fd):
    os.lseek(fd, 0, os.SEEK_SET)
    return str(os.read(fd, 64 * 1024), "utf-8")


def replaceFdContent(fd, content):
    os.lseek(fd, 0, os.SEEK_SET)
    os.ftruncate(fd, 0)
    os.write(fd, bytes(content, "utf-8"))


def appendFdContent(fd, content):
    os.lseek(fd, 0, os.SEEK_END)
    os.write(fd, bytes(content, "utf-8"))


DOCUMENT_ADD_FLAGS_REUSE_EXISTING = 1 << 0
DOCUMENT_ADD_FLAGS_PERSISTENT = 1 << 1
DOCUMENT_ADD_FLAGS_AS_NEEDED_BY_APP = 1 << 2
DOCUMENT_ADD_FLAGS_DIRECTORY = 1 << 3


def assertRaises(exc_type, func, *args, **kwargs):
    with pytest.raises(exc_type):
        func(*args, **kwargs)


def assertRaisesErrno(error_nr, func, *args, **kwargs):
    with pytest.raises(OSError) as excinfo:
        func(*args, **kwargs)

    if excinfo.value.errno != error_nr:
        raise AssertionError(
            "Wrong errno {0} was raised instead of {1}".format(
                excinfo.value.errno, error_nr
            )
        )


def assertRaisesGError(message, code, func, *args, **kwargs):
    with pytest.raises(GLib.GError) as excinfo:
        func(*args, **kwargs)

    if not excinfo.value.message.startswith(message):
        raise AssertionError(
            "Wrong message {0} doesn't start with {1}".format(
                excinfo.value.message, message
            )
        )
    if excinfo.value.code != code:
        raise AssertionError(
            "Wrong code {0} was raised instead of {1}".format(excinfo.value.code, code)
        )


def assertFileHasContent(path, expected_content):
    with open(path) as f:
        file_content = f.read()
        assert file_content == expected_content


def assertFdHasContent(fd, expected_content):
    content = readFdContent(fd)
    assert content == expected_content


def assertSameStat(a, b, b_mode_mask):
    if not (
        a.st_mode == (b.st_mode & b_mode_mask)
        and a.st_nlink == b.st_nlink
        and a.st_size == b.st_size
        and a.st_uid == b.st_uid
        and a.st_gid == b.st_gid
        and a.st_atime == b.st_atime
        and a.st_mtime == b.st_mtime
        and a.st_ctime == b.st_ctime
    ):
        raise AssertionError("Stat value {} was not the expected {})".format(a, b))


def assertFileExist(path):
    try:
        info = os.lstat(path)
        if info.st_mode & stat.S_IFREG != stat.S_IFREG:
            raise AssertionError("File {} is not a regular file".format(path))
    except Exception:
        raise AssertionError("File {} doesn't exist".format(path))


def assertDirExist(path):
    try:
        info = os.lstat(path)
        if info.st_mode & stat.S_IFDIR != stat.S_IFDIR:
            raise AssertionError("File {} is not a directory file".format(path))
    except Exception:
        raise AssertionError("File {} doesn't exist".format(path))


def assertSymlink(path, expected_target):
    try:
        info = os.lstat(path)
        if info.st_mode & stat.S_IFLNK != stat.S_IFLNK:
            raise AssertionError("File {} is not a symlink".format(path))
        target = os.readlink(path)
        if target != expected_target:
            raise AssertionError(
                "File {} has wrong target {}, expected {}".format(
                    path, target, expected_target
                )
            )
    except Exception:
        raise AssertionError("Symlink {} doesn't exist".format(path))


def assertFileNotExist(path):
    try:
        os.lstat(path)
    except FileNotFoundError:
        return
    except Exception:
        raise AssertionError(
            "Got wrong execption {} for {}, expected FileNotFoundError".format(
                sys.exc_info()[0], path
            )
        )
    raise AssertionError("Path {} unexpectedly exists".format(path))


def assertDirFiles(path, expected_files, exhaustive=True, volatile_files=None):
    found_files = os.listdir(path)
    remaining = set(found_files)
    for file in expected_files:
        if file in remaining:
            remaining.remove(file)
        elif file not in volatile_files:
            raise AssertionError(
                "Expected file {} not found in dir {} (all: {})".format(
                    file, path, found_files
                )
            )
    if exhaustive:
        if len(remaining) != 0:
            raise AssertionError(
                "Unexpected files {} in dir {} (all: {})".format(
                    remaining, path, found_files
                )
            )


class Doc:
    def __init__(self, portal, id, path, content, is_dir=False):
        self.portal = portal
        self.id = id
        self.content = content
        self.real_path = path
        self.is_dir = is_dir
        self.apps = []
        self.files = []

        if is_dir:
            self.real_dirname = path
            self.filename = None
            self.dirname = os.path.basename(path)
        else:
            (self.real_dirname, self.filename) = os.path.split(path)
            self.dirname = None
            if content:
                self.files.append(self.filename)

    def is_readable_by(self, app_id):
        if app_id:
            return app_id in self.apps
        return True

    def is_writable_by(self, app_id):
        if app_id:
            return app_id in self.apps and ".write." in app_id
        else:
            return True

    def get_doc_path(self, app_id):
        if app_id:
            base = self.portal.app_path(app_id) + "/" + self.id
        else:
            base = self.portal.mountpoint + "/" + self.id
        if self.is_dir:
            return base + "/" + self.dirname
        else:
            return base

    def __str__(self):
        name = self.id
        if self.is_dir:
            return "%s(dir)" % (name)
        elif self.content is None:
            return "%s(missing)" % (name)
        else:
            return "%s" % (name)


class DocPortal:
    def __init__(self):
        self.apps = []
        self.volatile_apps = set()
        self.docs = {}
        self.bus = Gio.bus_get_sync(Gio.BusType.SESSION, None)
        self.proxy = Gio.DBusProxy.new_sync(
            self.bus,
            Gio.DBusProxyFlags.NONE,
            None,
            "org.freedesktop.portal.Documents",
            "/org/freedesktop/portal/documents",
            "org.freedesktop.portal.Documents",
            None,
        )
        self.mountpoint = self.get_mount_path()

    def get_mount_path(self):
        res = self.proxy.call_sync("GetMountPoint", GLib.Variant("()", ()), 0, -1, None)
        return bytearray(res[0][:-1]).decode("utf-8")

    def grant_permissions(self, doc_id, app_id, permissions):
        self.proxy.call_sync(
            "GrantPermissions",
            GLib.Variant("(ssas)", (doc_id, app_id, permissions)),
            0,
            -1,
            None,
        )

    def lookup(self, path):
        res = self.proxy.call_sync(
            "Lookup", GLib.Variant("(ay)", (filename_to_ay(path),)), 0, -1, None
        )
        return res[0]

    def delete(self, doc_id):
        self.proxy.call_sync("Delete", GLib.Variant("(s)", (doc_id,)), 0, -1, None)
        del self.docs[doc_id]

    def add(self, path, reuse_existing=True):
        fdlist = Gio.UnixFDList.new()
        fd = os.open(path, os.O_PATH)
        handle = fdlist.append(fd)
        os.close(fd)
        res = self.proxy.call_with_unix_fd_list_sync(
            "Add",
            GLib.Variant("(hbb)", (handle, reuse_existing, False)),
            0,
            -1,
            fdlist,
            None,
        )
        doc_id = res[0][0]
        if doc_id in self.docs:
            return self.docs[doc_id]

        with open(path) as f:
            content = f.read()
        doc = Doc(self, doc_id, path, content)
        self.docs[doc.id] = doc
        return doc

    def add_named(self, path, reuse_existing=True):
        (dirname, filename) = os.path.split(path)
        fdlist = Gio.UnixFDList.new()
        fd = os.open(dirname, os.O_PATH)
        handle = fdlist.append(fd)
        os.close(fd)
        res = self.proxy.call_with_unix_fd_list_sync(
            "AddNamed",
            GLib.Variant(
                "(haybb)", (handle, filename_to_ay(filename), reuse_existing, False)
            ),
            0,
            -1,
            fdlist,
            None,
        )
        doc_id = res[0][0]
        if doc_id in self.docs:
            return self.docs[doc_id]

        try:
            with open(path) as f:
                content = f.read()
        except Exception:
            content = None
        doc = Doc(self, doc_id, path, content)
        self.docs[doc.id] = doc
        return doc

    def add_full(self, path, flags):
        fdlist = Gio.UnixFDList.new()
        fd = os.open(path, os.O_PATH)
        handle = fdlist.append(fd)
        os.close(fd)
        res = self.proxy.call_with_unix_fd_list_sync(
            "AddFull",
            GLib.Variant("(ahusas)", ([handle], flags, "", [])),
            0,
            -1,
            fdlist,
            None,
        )
        doc_id = res[0][0][0]
        if doc_id in self.docs:
            return self.docs[doc_id]
        doc = Doc(self, doc_id, path, True, (flags & DOCUMENT_ADD_FLAGS_DIRECTORY) != 0)
        self.docs[doc.id] = doc
        return doc

    def add_dir(self, path):
        return self.add_full(
            path, DOCUMENT_ADD_FLAGS_REUSE_EXISTING | DOCUMENT_ADD_FLAGS_DIRECTORY
        )

    def get_docs_for_app(self, app_id):
        docs = []
        for doc in self.docs.values():
            if doc.is_readable_by(app_id):
                docs.append(doc.id)
        return docs

    def ensure_app_id(self, app_id, volatile=False):
        if app_id not in self.apps:
            self.apps.append(app_id)
        if volatile:
            self.volatile_apps.add(app_id)

    def get_docs(self):
        return list(self.docs.values())

    def get_docs_randomized(self):
        docs = list(self.docs.values())
        random.shuffle(docs)
        return docs

    def get_doc(self, doc_id):
        return self.docs[doc_id]

    def get_app_ids(self):
        return self.apps

    def get_volatile_app_ids(self):
        return self.volatile_apps

    def get_app_ids_randomized(self):
        apps = self.apps.copy()
        random.shuffle(apps)
        return apps

    def by_app_path(self):
        return self.mountpoint + "/by-app"

    def app_path(self, app_id):
        return self.mountpoint + "/by-app/" + app_id


class FileTransferPortal(DocPortal):
    def __init__(self):
        super().__init__()
        self.ft_proxy = Gio.DBusProxy.new_sync(
            self.bus,
            Gio.DBusProxyFlags.NONE,
            None,
            "org.freedesktop.portal.Documents",
            "/org/freedesktop/portal/documents",
            "org.freedesktop.portal.FileTransfer",
            None,
        )

    def start_transfer(self):
        res = self.ft_proxy.call_sync(
            "StartTransfer", GLib.Variant("(a{sv})", ([None])), 0, -1, None
        )
        return res[0]

    def add_files(self, key, files):
        fdlist = Gio.UnixFDList.new()
        handles = []
        for filename in files:
            fd = os.open(filename, os.O_PATH)
            handle = fdlist.append(fd)
            handles.append(handle)
            os.close(fd)

        res = self.ft_proxy.call_with_unix_fd_list_sync(
            "AddFiles",
            GLib.Variant("(saha{sv})", (key, handles, [])),
            0,
            -1,
            fdlist,
            None,
        )
        return res

    def retrieve_files(self, key):
        res = self.ft_proxy.call_sync(
            "RetrieveFiles",
            GLib.Variant("(sa{sv})", (key, [])),
            0,
            -1,
            None,
        )
        return res

    def stop_transfer(self, key):
        res = self.ft_proxy.call_sync(
            "StopTransfer",
            GLib.Variant("(s)", (key,)),
            0,
            -1,
            None,
        )
        return res


def check_virtual_stat(info, writable=False):
    assert info.st_uid == os.getuid()
    assert info.st_gid == os.getgid()
    if writable:
        assert info.st_mode == stat.S_IFDIR | 0o700
    else:
        assert info.st_mode == stat.S_IFDIR | 0o500


def verify_virtual_dir(path, files, volatile_files=None):
    info = os.lstat(path)
    check_virtual_stat(info)
    assert os.access(path, os.R_OK)
    assert not os.access(path, os.W_OK)

    assertRaises(FileNotFoundError, os.lstat, path + "/not-existing-file")

    if files is not None:
        assertDirFiles(path, files, ensure_no_remaining, volatile_files)


def verify_doc(doc, app_id=None):
    dir = doc.get_doc_path(app_id)

    if doc.is_dir:
        vdir = os.path.dirname(dir)
        info = os.lstat(vdir)
        check_virtual_stat(info)
        pass
    else:
        info = os.lstat(dir)
        check_virtual_stat(info, doc.is_writable_by(app_id))
        assert os.access(dir, os.R_OK)
        if doc.is_writable_by(app_id):
            assert os.access(dir, os.W_OK)
        else:
            assert not os.access(dir, os.W_OK)

    assertRaises(FileNotFoundError, os.lstat, dir + "/not-existing-file")

    assertDirFiles(dir, doc.files)

    for file in doc.files:
        filepath = dir + "/" + file
        info = os.lstat(filepath)
        assert info.st_uid == os.getuid()
        assert info.st_gid == os.getgid()

        assert os.access(filepath, os.R_OK)
        if doc.is_writable_by(app_id):
            assert os.access(filepath, os.W_OK)
        else:
            assert not os.access(filepath, os.W_OK)

    if doc.filename:
        main_path = dir + "/" + doc.filename
        real_path = doc.real_path
        if doc.content:
            assertFileExist(main_path)
            assertFileHasContent(main_path, doc.content)
            assertFileHasContent(real_path, doc.content)

            info = os.lstat(main_path)
            real_info = os.lstat(real_path)
            mode_mask = ~(stat.S_ISUID | stat.S_ISGID | stat.S_ISVTX)
            if not doc.is_writable_by(app_id):
                mode_mask = mode_mask & ~(stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH)
            assertSameStat(info, real_info, mode_mask)

        else:
            assertRaises(FileNotFoundError, os.lstat, main_path)
            assertRaises(FileNotFoundError, os.open, main_path, os.O_RDONLY)
            assertRaises(FileNotFoundError, os.lstat, doc.real_path)
            assertRaises(FileNotFoundError, os.open, doc.real_path, os.O_RDONLY)

    # Ensure no leftover temp files
    for real_file in os.listdir(os.path.dirname(doc.real_path)):
        assert not real_file.startswith(".xdp")


def verify_fs_layout(portal):
    verify_virtual_dir(portal.mountpoint, ["by-app"] + list(portal.docs.keys()))
    verify_virtual_dir(
        portal.by_app_path(), portal.get_app_ids(), portal.get_volatile_app_ids()
    )

    for doc in portal.get_docs():
        verify_doc(doc)

    # Verify the by-app subdirs (just the directory for now)
    for app_id in portal.get_app_ids():
        docs_for_app = portal.get_docs_for_app(app_id)
        verify_virtual_dir(portal.app_path(app_id), docs_for_app)
        for doc_id in docs_for_app:
            doc = portal.get_doc(doc_id)
            verify_doc(doc, app_id)


def check_virtdir_perms(path):
    assertRaises(PermissionError, os.mkdir, path + "/a_dir")
    assertRaises(PermissionError, os.open, path + "/a-file", os.O_RDWR | os.O_CREAT)


def check_root_perms(path):
    check_virtdir_perms(path)
    assertRaises(PermissionError, os.rename, path + "/by-app", path + "/by-app2")
    assertRaises(PermissionError, os.rmdir, path + "/by-app")


def check_byapp_perms(path):
    check_virtdir_perms(path)
    assertRaises(PermissionError, os.mkdir, path + "/a_dir")


def check_regular_doc_perms(doc, app_id):
    path = doc.get_doc_path(app_id)
    writable = doc.is_writable_by(app_id)
    # regular documents, can't do most stuff
    assertRaises(PermissionError, os.mkdir, path + "/dir")
    assertRaises(PermissionError, os.symlink, "symlink-value", path + "/symlink")

    docpath = path + "/" + doc.filename
    tmppath = path + "/a-tmpfile"
    tmppath2 = path + "/another-tmpfile"
    if doc.content:  # Main file exists
        assertFileExist(docpath)
        assertFileExist(doc.real_path)
        assertRaises(PermissionError, os.link, docpath, path + "/a-hardlink")
        assertRaises(NotADirectoryError, os.rmdir, docpath)
        assertRaises(PermissionError, os.setxattr, docpath, "user.attr", b"foo")
        assertRaises(PermissionError, os.removexattr, docpath, "user.attr")

        fd = os.open(docpath, os.O_RDONLY, 0o600)
        os.close(fd)

        if not writable:
            assertRaises(
                PermissionError, os.open, docpath, os.O_RDONLY | os.O_TRUNC, 0o600
            )
            assertRaises(PermissionError, os.open, docpath, os.O_WRONLY, 0o600)
            assertRaises(PermissionError, os.open, docpath, os.O_RDWR, 0o600)
            assertRaises(PermissionError, os.rename, docpath, docpath + "renamed")
            assertRaises(PermissionError, os.truncate, docpath, 1)
            assertRaises(PermissionError, os.unlink, docpath)
            assertRaises(PermissionError, os.utime, docpath)
        else:
            # Can't move file out of docdir or into other version of same docdir
            assertRaisesErrno(
                errno.EXDEV, os.rename, docpath, path + "/../" + doc.filename
            )
            if app_id:
                assertRaisesErrno(
                    errno.EXDEV,
                    os.rename,
                    docpath,
                    doc.get_doc_path(None) + doc.filename,
                )
            if doc.apps and app_id != doc.apps[0]:
                assertRaisesErrno(
                    errno.EXDEV,
                    os.rename,
                    docpath,
                    doc.get_doc_path(doc.apps[0]) + doc.filename,
                )

            # Ensure we can read it (multiple times)
            fd = os.open(docpath, os.O_RDONLY, 0o600)
            assertFdHasContent(fd, doc.content)
            assertFdHasContent(fd, doc.content)

            # Ensure we can rename it
            os.rename(docpath, docpath + "_renamed")
            assertRaises(FileNotFoundError, os.open, docpath, os.O_RDONLY, 0o600)
            # ... and still read it
            assertFdHasContent(fd, doc.content)

            # Ensure we can delete it
            os.unlink(docpath + "_renamed")
            # ... and still read it
            assertFdHasContent(fd, doc.content)
            os.close(fd)

            # Replace main file with rename of tmpfile
            setFileContent(docpath, "orig-data")
            fd1 = os.open(docpath, os.O_RDONLY, 0o600)

            setFileContent(tmppath, "new-data")
            fd2 = os.open(tmppath, os.O_RDONLY, 0o600)

            os.rename(tmppath, tmppath2)
            assertRaises(FileNotFoundError, os.lstat, tmppath)
            assertFdHasContent(fd2, "new-data")
            assertFileHasContent(tmppath2, "new-data")

            os.rename(tmppath2, docpath)
            assertRaises(FileNotFoundError, os.lstat, tmppath2)
            assertFdHasContent(fd1, "orig-data")
            assertFdHasContent(fd2, "new-data")
            assertFileHasContent(docpath, "new-data")
            appendFileContent(docpath, "-more")
            assertFdHasContent(fd2, "new-data-more")

            setFileContent(tmppath, "replace-this-data")
            fd3 = os.open(tmppath, os.O_RDONLY, 0o600)
            os.rename(docpath, tmppath)
            assertFdHasContent(fd2, "new-data-more")
            assertFdHasContent(fd3, "replace-this-data")
            fd4 = os.open(tmppath, os.O_RDWR, 0o600)
            assertFdHasContent(fd4, "new-data-more")

            # Restore original version
            os.rename(tmppath, docpath)
            replaceFdContent(fd4, doc.content)
            assertFdHasContent(fd2, doc.content)
            assertFdHasContent(fd4, doc.content)
            assertFileHasContent(docpath, doc.content)
            assertFdHasContent(fd1, "orig-data")
            assertFdHasContent(fd3, "replace-this-data")

            os.close(fd1)
            os.close(fd2)
            os.close(fd3)
            os.close(fd4)

            assertRaises(NotADirectoryError, os.rmdir, docpath)
            assertRaises(PermissionError, os.link, docpath, path + "/a-hardlink")
            assertRaises(PermissionError, os.setxattr, docpath, "user.attr", b"foo")
            assertRaises(PermissionError, os.removexattr, docpath, "user.attr")

    else:  # Main file doesn't exist
        assertFileNotExist(docpath)
        assertFileNotExist(doc.real_path)
        if writable:  # But we can create it
            setFileContent(docpath, "some-data")
            assertFileHasContent(docpath, "some-data")
            os.unlink(docpath)
        else:  # And we can't create it
            assertRaises(
                PermissionError,
                os.open,
                docpath,
                os.O_CREAT | os.O_RDONLY | os.O_TRUNC,
                0o600,
            )
            assertRaises(
                PermissionError, os.open, docpath, os.O_CREAT | os.O_WRONLY, 0o600
            )
            assertRaises(
                PermissionError, os.open, docpath, os.O_CREAT | os.O_RDWR, 0o600
            )

        # Ensure it show up if created outside
        setFileContent(doc.real_path, "from-outside")
        assertFileExist(docpath)
        assertFileHasContent(docpath, "from-outside")
        if writable:
            os.unlink(docpath)
        else:
            assertRaises(PermissionError, os.unlink, docpath)
            os.unlink(doc.real_path)
        assertFileNotExist(docpath)

    if writable:  # We can create tempfiles, do some simple checks
        setFileContent(tmppath, "tempdata")
        assertFileHasContent(tmppath, "tempdata")
        assertRaises(NotADirectoryError, os.rmdir, tmppath)
        assertRaises(PermissionError, os.link, tmppath, path + "/a-hardlink")
        assertRaises(PermissionError, os.setxattr, tmppath, "user.attr", b"foo")
        assertRaises(PermissionError, os.removexattr, tmppath, "user.attr")

        os.rename(tmppath, tmppath2)
        assertFileHasContent(tmppath2, "tempdata")
        os.unlink(tmppath2)
    else:
        # We should be unable to create tempfiles
        assertRaises(
            PermissionError,
            os.open,
            tmppath,
            os.O_CREAT | os.O_RDONLY | os.O_TRUNC,
            0o600,
        )
        assertRaises(PermissionError, os.open, tmppath, os.O_CREAT | os.O_WRONLY, 0o600)
        assertRaises(PermissionError, os.open, tmppath, os.O_CREAT | os.O_RDWR, 0o600)


def check_directory_doc_perms(doc, app_id):
    writable = doc.is_writable_by(app_id)

    docpath = doc.get_doc_path(app_id)
    realpath = doc.real_path

    # We should not be able to do anything with the toplevel document dir (other than reading the real dir)
    vpath = os.path.dirname(docpath)
    assertDirExist(vpath)
    assertDirFiles(vpath, [doc.dirname])
    assertRaises(PermissionError, os.mkdir, vpath + "/a_dir")
    assertRaises(PermissionError, os.rename, docpath, vpath + "/foo")
    assertRaises(PermissionError, os.rmdir, docpath)
    assertRaises(
        PermissionError, os.open, vpath + "/a_file", os.O_CREAT | os.O_RDWR, 0o600
    )

    assertDirExist(docpath)

    # Create some pre-existing files:

    real_dir = realpath + "/dir"
    os.mkdir(real_dir)
    setFileContent(real_dir + "/realfile", "real1")
    setFileContent(real_dir + "/readonly", "readonly")
    os.chmod(real_dir + "/readonly", 0o500)
    os.mkdir(real_dir + "/subdir")
    os.link(real_dir + "/realfile", real_dir + "/subdir/hardlink")
    os.symlink("realfile", real_dir + "/symlink")
    os.symlink("the-void", real_dir + "/broken-symlink")

    # Ensure they are visible via portal

    dir = docpath + "/dir"
    assertDirFiles(docpath, ["dir"])
    assertDirExist(dir)
    assertDirFiles(dir, ["realfile", "readonly", "subdir", "symlink", "broken-symlink"])
    assertDirExist(dir + "/subdir")
    assertDirFiles(dir + "/subdir", ["hardlink"])
    assertFileHasContent(dir + "/realfile", "real1")
    assertFileHasContent(dir + "/readonly", "readonly")
    assertFileHasContent(dir + "/subdir/hardlink", "real1")
    assert (
        os.lstat(dir + "/realfile").st_ino == os.lstat(dir + "/subdir/hardlink").st_ino
    )
    assertSymlink(dir + "/symlink", "realfile")
    assertSymlink(dir + "/broken-symlink", "the-void")

    filepath = docpath + "/a-file"
    real_filepath = doc.real_path + "/a-file"
    filepath2 = docpath + "/dir/a-file2"
    real_filepath2 = doc.real_path + "/dir/a-file2"

    if writable:  # We can create files
        if os.environ.get("TEST_IN_ROOTED_CI"):
            assertRaises(PermissionError, os.open, dir + "/readonly", os.O_RDWR)
            os.chmod(dir + "/readonly", 0o700)
            fd = os.open(dir + "/readonly", os.O_RDWR)  # Works now
            os.close(fd)

        setFileContent(filepath, "filedata")
        assertFileHasContent(filepath, "filedata")
        assertFileHasContent(real_filepath, "filedata")

        fd = os.open(filepath, os.O_RDONLY)
        fd2 = os.open(filepath, os.O_RDWR)
        assertFdHasContent(fd, "filedata")
        assertFdHasContent(fd2, "filedata")
        appendFdContent(fd2, "-more")
        assertFdHasContent(fd, "filedata-more")
        assertFdHasContent(fd2, "filedata-more")

        os.link(filepath, filepath2)
        assert os.lstat(filepath).st_ino == os.lstat(filepath2).st_ino
        assert os.lstat(filepath).st_ino == os.fstat(fd).st_ino
        assertFileHasContent(filepath2, "filedata-more")
        assertFileHasContent(real_filepath2, "filedata-more")

        os.unlink(filepath)
        assertFileNotExist(filepath)
        assertFileNotExist(real_filepath)
        assertFdHasContent(fd, "filedata-more")
        assertFdHasContent(fd2, "filedata-more")

        replaceFdContent(fd2, "replaced")
        assertFileHasContent(filepath2, "replaced")
        assertFileHasContent(real_filepath2, "replaced")
        assertFileNotExist(filepath)
        assertFileNotExist(real_filepath)
        assertFdHasContent(fd, "replaced")
        assertFdHasContent(fd2, "replaced")

        # Move between dirs
        os.rename(filepath2, docpath + "/moved")
        assertFileHasContent(docpath + "/moved", "replaced")

        assertRaisesErrno(errno.EXDEV, os.rename, docpath, doc.portal.mountpoint)

        os.unlink(docpath + "/moved")

        os.close(fd)
        os.close(fd2)

        os.symlink("realfile", dir + "/symlink2")
        os.symlink("the-void", dir + "/broken-symlink2")
        assertSymlink(dir + "/symlink2", "realfile")
        assertSymlink(dir + "/broken-symlink2", "the-void")
        os.unlink(dir + "/symlink2")
        os.unlink(dir + "/broken-symlink2")

    else:
        # We should be unable to create files
        assertRaises(
            PermissionError,
            os.open,
            filepath,
            os.O_CREAT | os.O_RDONLY | os.O_TRUNC,
            0o600,
        )
        assertRaises(
            PermissionError, os.open, filepath, os.O_CREAT | os.O_WRONLY, 0o600
        )
        assertRaises(PermissionError, os.open, filepath, os.O_CREAT | os.O_RDWR, 0o600)

        assertRaises(PermissionError, os.open, dir + "/realfile", os.O_RDWR)
        assertRaises(PermissionError, os.open, dir + "/readonly", os.O_RDWR)
        assertRaises(PermissionError, os.truncate, dir + "/realfile", 0)
        assertRaises(PermissionError, os.link, dir + "/realfile", dir + "/foo")
        assertRaises(PermissionError, os.symlink, "foo", dir + "/new-symlink")
        assertRaises(PermissionError, os.rename, dir + "/realfile", dir + "/foo")
        assertRaises(PermissionError, os.unlink, dir + "/realfile")
        assertRaises(PermissionError, os.chmod, dir + "/realfile", 0o700)
        assertRaises(PermissionError, os.rmdir, dir + "/subdir")

    os.unlink(real_dir + "/realfile")
    os.unlink(real_dir + "/readonly")
    os.unlink(real_dir + "/subdir/hardlink")
    os.unlink(real_dir + "/symlink")
    os.unlink(real_dir + "/broken-symlink")
    os.rmdir(real_dir + "/subdir")
    os.rmdir(real_dir)


def check_doc_perms(doc, app_id):
    path = doc.get_doc_path(app_id)
    readable = doc.is_readable_by(app_id)
    if not readable:
        assertRaises(FileNotFoundError, os.lstat, path)
        if doc.is_dir:  # Non readable dir means we can't even see the toplevel dir
            assertRaises(FileNotFoundError, os.mkdir, path)
        else:
            assertRaises(PermissionError, os.mkdir, path)
        return

    assertRaises(PermissionError, os.rmdir, path)
    assertRaises(PermissionError, os.rename, path, path + "_renamed")
    assertRaises(IsADirectoryError, os.unlink, path)
    if doc.is_dir:
        check_directory_doc_perms(doc, app_id)
    else:
        check_regular_doc_perms(doc, app_id)


def check_perms(portal):
    check_root_perms(portal.mountpoint)
    check_byapp_perms(portal.by_app_path())

    for doc in portal.get_docs_randomized():
        check_doc_perms(doc, None)
        for app_id in portal.get_app_ids_randomized():
            check_doc_perms(doc, app_id)


# Ensure that a single lookup by app-id creates that app id (we need this for when mounting the subdir for an app)
def create_app_by_lookup(portal):
    # Should only work for valid app ids
    assertRaises(FileNotFoundError, os.lstat, portal.app_path("not-an-app-id"))

    app_id = app_prefix + "Lookup"
    info = os.lstat(portal.app_path(app_id))
    check_virtual_stat(info)
    portal.ensure_app_id(app_id, volatile=True)


def ensure_real_dir(create_hidden_file=True):
    count = get_a_count("doc")
    dir = os.environ["TMPDIR"] + "/" + dir_prefix + str(count)
    os.makedirs(dir)
    if create_hidden_file:
        setFileContent(dir + "/cant-see-this-file", "s3krit")
    return (dir, count)


def ensure_real_dir_file(create_file):
    (dir, count) = ensure_real_dir()
    path = dir + "/the-file"
    if create_file:
        setFileContent(path, "data" + str(count))
    return path


def export_a_doc(portal):
    path = ensure_real_dir_file(True)
    doc = portal.add(path)
    logv("exported %s as %s" % (path, doc))

    lookup = portal.lookup(path)
    assert lookup == doc.id

    lookup_on_fuse = portal.lookup(doc.get_doc_path(None) + "/" + doc.filename)
    assert lookup_on_fuse == doc.id

    reused_doc = portal.add(path)
    assert doc is reused_doc

    not_reused_doc = portal.add(path, False)
    assert doc is not not_reused_doc

    # We should not be able to re-export a tmpfile
    tmppath = doc.get_doc_path(None) + "/tmpfile"
    setFileContent(tmppath, "tempdata")

    # Should not be able to add a tempfile on the fuse mount, or look it up
    assertRaises(GLib.Error, portal.add, tmppath)
    lookup = portal.lookup(tmppath)
    assert lookup == ""

    os.unlink(tmppath)


def export_a_named_doc(portal, create_file):
    path = ensure_real_dir_file(create_file)
    doc = portal.add_named(path)
    logv("exported (named) %s as %s" % (path, doc))

    if create_file:
        lookup = portal.lookup(path)
        assert lookup == doc.id

    reused_doc = portal.add_named(path)
    assert doc is reused_doc

    not_reused_doc = portal.add_named(path, False)
    assert doc is not not_reused_doc


def export_a_dir_doc(portal):
    (dir, count) = ensure_real_dir(False)
    doc = portal.add_dir(dir)
    logv("exported (dir) %s as %s" % (dir, doc))

    lookup = portal.lookup(dir)
    assert lookup == doc.id

    lookup_on_fuse = portal.lookup(doc.get_doc_path(None))
    assert lookup_on_fuse == doc.id

    # We should not be able to portal lookup a file in the dir doc
    subpath = doc.get_doc_path(None) + "/sub"
    setFileContent(subpath, "sub")
    doc = portal.lookup(subpath)
    assert doc == ""
    doc2 = portal.lookup(dir + "/sub")
    assert doc2 == ""

    # But we should be able to re-export the file
    reexported_doc = portal.add(subpath)
    reexported_docdir = reexported_doc.get_doc_path(None)
    assertFileHasContent(reexported_docdir + "/sub", "sub")
    portal.delete(reexported_doc.id)

    os.unlink(subpath)

    # And also re-export a directory
    os.mkdir(subpath)
    setFileContent(subpath + "/subfile", "subfile")
    reexported_doc = portal.add_dir(subpath)
    reexported_docdir = reexported_doc.get_doc_path(None)
    assertFileHasContent(reexported_docdir + "/subfile", "subfile")
    portal.delete(reexported_doc.id)

    os.unlink(subpath + "/subfile")
    os.rmdir(subpath)


def add_an_app(portal, num_docs):
    if num_docs == 0:
        return
    count = get_a_count("app")
    read_app = app_prefix + "read.App" + str(count)
    write_app = app_prefix + "write.App" + str(count)
    portal.ensure_app_id(read_app)
    portal.ensure_app_id(write_app)

    docs = portal.get_docs()
    ids = []
    for i in range(num_docs):
        if len(docs) == 0:
            continue
        indx = random.randint(0, len(docs) - 1)
        doc = docs[indx]
        del docs[indx]
        ids.append(doc.id)
        portal.grant_permissions(doc.id, read_app, ["read"])
        doc.apps.append(read_app)
        portal.grant_permissions(doc.id, write_app, ["read", "write"])
        doc.apps.append(write_app)
    logv("granted acces to %s and %s for %s" % (read_app, write_app, ids))


def file_transfer_portal_test():
    log("File transfer tests")
    ft_portal = FileTransferPortal()

    key = ft_portal.start_transfer()

    file1 = ensure_real_dir_file(True)
    file2 = ensure_real_dir_file(True)
    res = ft_portal.add_files(key, [file1, file2])

    res = ft_portal.retrieve_files(key)
    files = res[0]
    assert len(files) == 2
    # This is the same app, it's not sandboxed
    assert files[0] == file1
    assert files[1] == file2
    log("filetransfer tests ok")

    log("filetransfer dir")
    key = ft_portal.start_transfer()
    dir1 = ensure_real_dir(True)
    ft_portal.add_files(key, [file1, dir1[0], file2])
    res = ft_portal.retrieve_files(key)
    assert len(res[0]) == 3
    assert res[0][0] == file1
    assert res[0][1] == dir1[0]
    assert res[0][2] == file2
    log("filetransfer dir ok")

    log("filetransfer key")
    # Test that an invalid key is rejected
    key = ft_portal.start_transfer()
    assert key != "1234"
    assertRaisesGError(
        "GDBus.Error:org.freedesktop.DBus.Error.AccessDenied",
        9,
        ft_portal.add_files,
        "1234",
        [file1, file2],
    )

    # Test stop transfer
    key = ft_portal.start_transfer()
    ft_portal.add_files(key, [file1, file2])
    ft_portal.stop_transfer(key)
    assertRaisesGError(
        "GDBus.Error:org.freedesktop.DBus.Error.AccessDenied",
        9,
        ft_portal.retrieve_files,
        key,
    )
    assertRaisesGError(
        "GDBus.Error:org.freedesktop.DBus.Error.AccessDenied",
        9,
        ft_portal.add_files,
        key,
        [file1, file2],
    )

    # Test that we can't reuse an old key
    new_key = ft_portal.start_transfer()
    assertRaisesGError(
        "GDBus.Error:org.freedesktop.DBus.Error.AccessDenied",
        9,
        ft_portal.add_files,
        key,
        [file1, file2],
    )
    res = ft_portal.add_files(new_key, [file1, file2])
    log("filetransfer key ok")

    log("File transfer tests ok")


def run_test(iterations, prefix=None, do_ensure_no_remaining=True):
    global app_prefix
    global dir_prefix
    global ensure_no_remaining

    if prefix:
        app_prefix = app_prefix + prefix + "."
        dir_prefix = dir_prefix + "-" + prefix + "-"

    ensure_no_remaining = do_ensure_no_remaining

    log("Connecting to portal")
    doc_portal = DocPortal()

    log("Running fuse tests...")
    create_app_by_lookup(doc_portal)
    verify_fs_layout(doc_portal)

    log("Creating some docs")
    for i in range(10):
        export_a_doc(doc_portal)
        verify_fs_layout(doc_portal)

    log("Creating some named docs (existing)")
    for i in range(10):
        export_a_named_doc(doc_portal, True)
    verify_fs_layout(doc_portal)

    log("Creating some named docs (non-existing)")
    for i in range(10):
        export_a_named_doc(doc_portal, False)
    verify_fs_layout(doc_portal)

    log("Creating some dir docs")
    for i in range(10):
        export_a_dir_doc(doc_portal)
    verify_fs_layout(doc_portal)

    log("Creating some apps")
    for i in range(10):
        add_an_app(doc_portal, 6)
    verify_fs_layout(doc_portal)

    for i in range(iterations):
        log("Checking permissions, pass %d" % (i + 1))
        check_perms(doc_portal)
        verify_fs_layout(doc_portal)

    log("fuse tests ok")
    file_transfer_portal_test()


class Process(mp.Process):
    def __init__(self, *args, **kwargs):
        mp.Process.__init__(self, *args, **kwargs)
        self._pconn, self._cconn = mp.Pipe()
        self._exception = None

    def run(self):
        try:
            mp.Process.run(self)
            self._cconn.send(None)
        except Exception as e:
            tb = traceback.format_exc()
            self._cconn.send((e, tb))

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception


class TestDocumentFuse:
    def parallel(self, test_function, parallel_tests, parallel_iterations):
        procs = []
        for i in range(parallel_tests):
            p = Process(
                target=test_function, args=(parallel_iterations, f"c{i}", False)
            )
            p.start()
            procs.append(p)

        for p in procs:
            p.join()

            if p.exception:
                error, traceback = p.exception
                raise error

    def test_single_thread(self, portals, xdg_document_portal, dbus_con):
        run_test(3)

    def test_multi_thread(self, portals, xdg_document_portal, dbus_con):
        if xdp.run_long_tests():
            return self.parallel(run_test, 20, 10)
        if xdp.is_in_ci():
            return self.parallel(run_test, 5, 3)
        self.parallel(run_test, 10, 5)


# Running
# ./tests/run-test.sh -n 0 tests/test_document_fuse.py::TestDocumentFuse::test_multi_thread
# works fine, but with
# ./tests/run-test.sh -n 0 tests/test_document_fuse.py::TestDocumentFuse
# the `test_multi_thread` test is failing.
# For now, let's skip the test and turn it on again when we have fixed it.
pytest.skip("Test has a race condition which can make it fail", allow_module_level=True)

try:
    xdp.ensure_fuse_supported()
except xdp.FuseNotSupportedException as e:
    pytest.skip(f"No fuse support: {e}", allow_module_level=True)



# SPDX-License-Identifier: LGPL-2.1-or-later
#
# This file is formatted with Python Black

import os
from pathlib import Path
from gi.repository import GLib, Gio


EXPORT_FILES_FLAG_EXPORT_DIR = 8


def path_from_null_term_bytes(bytes):
    path_bytes, rest = bytes.split(b"\x00")
    assert rest == b""
    return Path(os.fsdecode(path_bytes))


def get_mountpoint(documents_intf):
    mountpoint = documents_intf.GetMountPoint(byte_arrays=True)
    mountpoint = path_from_null_term_bytes(mountpoint)
    assert mountpoint.exists()
    return mountpoint


def export_file(documents_intf, file_path, unique=False):
    assert file_path.exists()

    with open(file_path.absolute().as_posix(), "r") as file:
        doc_id = documents_intf.Add(file.fileno(), not unique, False)
        assert doc_id

    return doc_id


def export_file_named(documents_intf, folder_path, name, unique=False):
    assert folder_path.exists()

    # bytestring convention is zero terminated
    name_nt = os.fsencode(name) + b"\x00"

    try:
        fd = os.open(folder_path.absolute().as_posix(), os.O_PATH | os.O_CLOEXEC)
        doc_id = documents_intf.AddNamed(fd, name_nt, not unique, False)
        assert doc_id
    finally:
        os.close(fd)

    return doc_id


def export_files(documents_intf, file_paths, perms, flags=0, app_id=""):
    fds = []
    try:
        for file_path in file_paths:
            fds.append(
                os.open(file_path.absolute().as_posix(), os.O_PATH | os.O_CLOEXEC)
            )

        result = documents_intf.AddFull(
            fds,
            flags,
            app_id,
            perms,
            byte_arrays=True,
        )
    finally:
        for fd in fds:
            os.close(fd)

    assert result
    return result


def write_bytes_atomic(file_path, bytes):
    GLib.file_set_contents(file_path.absolute().as_posix(), bytes)


def write_bytes_trunc(file_path, bytes):
    try:
        fd = os.open(
            file_path.absolute().as_posix(), os.O_RDWR | os.O_TRUNC | os.O_CREAT
        )
        os.write(fd, bytes)
    finally:
        os.close(fd)


def get_host_path_attr(path):
    xattr = "xattr::document-portal.host-path"
    file = Gio.file_new_for_path(path.absolute().as_posix())
    info = file.query_info(xattr, Gio.FileQueryInfoFlags.NONE)
    host_path = info.get_attribute_as_string(xattr)
    if not host_path:
        return None
    return Path(os.fsdecode(host_path))
